"""
You have built a custom container object that internally holds a list,
tuple or some other iterable. You would like to make iteration work
with your new container.
"""


class Node:
    def __init__(self, value):
        self.value = value
        self._children = []

    def __repr__(self):
        return f"Node({self.value})"

    def add_child(self, node):
        self._children.append(node)

    def __iter__(self):
        return iter(self._children)


def main():
    root = Node(0)
    child1 = Node(1)
    child2 = Node(2)
    root.add_child(child1)
    root.add_child(child2)

    print(f"root node: {root}")

    for i, child in enumerate(root, 1):
        print(f"Child {i}: {child}")


if __name__ == "__main__":
    main()



from pathlib import Path
from .errors import RecipeNotFoundError
from .recipe import Recipe
from .utils.text import clean_text, extract_leading_numbers


class Chapter:
    def __init__(self, chapter_dir: Path):
        self.chapter_dir = chapter_dir
        self.recipes: dict[int, Recipe] = {}
        self._collect()

    @property
    def number(self):
        return extract_leading_numbers(self.chapter_dir.stem)

    @property
    def name(self):
        return self.chapter_dir.stem

    @property
    def size(self):
        return len(self.recipes)

    def search(self, pattern: str, ignore_case: bool = False):
        results = {}
        for recipe in self.recipes.values():
            matches = recipe.search(pattern, ignore_case)
            if matches:
                results.setdefault(recipe, []).extend(matches)
        return results

    def _collect(self):
        """Collects recipes"""
        for recipe_dir in sorted(self.chapter_dir.glob("[0-9]*")):
            recipe = Recipe(recipe_dir)
            self.recipes[recipe.number] = recipe

    def __getitem__(self, key):
        recipe = self.recipes.get(key)
        if recipe:
            return recipe
        raise RecipeNotFoundError(f"{self.number}.{key}")

    def __iter__(self):
        for recipe in self.recipes.values():
            yield recipe

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(number={self.number}, name={self.name}, recipes={len(self.recipes)})"

    def __str__(self) -> str:
        return f"{clean_text(self.name).title()}"



"""
You want to define a decorator inside a class deinition and apply it to
other functions or methods.
"""

from functools import wraps


class A:
    # Decorator as an instance method
    def decorator1(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            print("Decorator 1")
            return func(*args, **kwargs)

        return wrapper

    # Decorator as a class method
    @classmethod
    def decorator2(cls, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            print("Decorator 2")
            return func(*args, **kwargs)

        return wrapper


def main():

    # As an instance method
    a = A()

    @a.decorator1
    def spam():
        print("Spam!")

    # As a class method
    @A.decorator2
    def grok():
        print("Grok!")

    spam()
    grok()


if __name__ == "__main__":
    main()



"""
You want to sarch for and replace a text pattern in a string.
"""

import re
from calendar import month_abbr


def change_date(m):
    mon_name = month_abbr[int(m.group(1))]
    return "{} {} {}".format(m.group(2), mon_name, m.group(3))


def main():
    # Some sample text
    text = "Today is 11/27/2012. PyCon starts 3/13/2013."

    datepat = re.compile(r"(\d+)/(\d+)/(\d+)")

    # (a) Simple substitution
    print(datepat.sub(r"\3-\1-\2", text))

    # (b) Replacement function
    print(datepat.sub(change_date, text))


if __name__ == "__main__":
    main()



import pytest
from pathlib import Path
from pyrecipes.recipe import Recipe, SearchMatch


def test_recipe__valid__exists(valid_recipe):
    assert valid_recipe.module == "example"
    assert valid_recipe.chapter == 1
    assert valid_recipe.chapter_name == "01_test_chapter"
    assert valid_recipe.number == 1
    assert valid_recipe.name == "01_test_recipe"
    assert valid_recipe.package == "pyrecipes.recipes0"
    assert valid_recipe.clean_name == "1) Test recipe"
    assert str(valid_recipe) == "1.1) Test recipe"
    assert valid_recipe.exists()


def test_recipe__valid__doesnt_exist(capsys):
    recipe = Recipe(Path("./01_testing_chapter/02_testing_example/"))
    recipe.run()
    out, err = capsys.readouterr()

    assert recipe.module == "example"
    assert recipe.chapter == 1
    assert recipe.chapter_name == "01_testing_chapter"
    assert recipe.number == 2
    assert recipe.name == "02_testing_example"
    assert recipe.package is None
    assert not recipe.exists()
    assert "Couldn't find Recipe" in out
    assert err == ""
    assert recipe.__repr__() == "Recipe(chapter=1, number=2)"
    with pytest.raises(ModuleNotFoundError, match="This recipe couldn't be found:"):
        recipe.get_module()


def test_recipes_running(recipe_path, capsys):
    recipe = Recipe(recipe_path)
    recipe.run()

    out, err = capsys.readouterr()
    assert "TODO" not in recipe.get_docstring()
    assert 'if __name__ == "__main__":' in recipe.get_code()
    assert f"Running {recipe.chapter}.{recipe.number}" in out
    assert err == ""


def test_recipes_search_found(valid_recipe):
    assert valid_recipe.search("TODO") == [
        SearchMatch(
            line_number=2,
            line_text="TODO\n",
            chapter=1,
            recipe_number=1,
            recipe_name="1) Test recipe",
            count=1,
        )
    ]


def test_recipes_search_ignore_case(valid_recipe):
    assert valid_recipe.search("todo", ignore_case=True) == [
        SearchMatch(
            line_number=2,
            line_text="TODO\n",
            chapter=1,
            recipe_number=1,
            recipe_name="1) Test recipe",
            count=1,
        )
    ]


def test_recipes_search__not_found(valid_recipe):
    assert valid_recipe.search("test") == []


def test_SearchMatches():
    matches = SearchMatch(1, "test", 2, 3, "test recipe", 1)
    assert matches.line_number == 1
    assert matches.line_text == "test"
    assert matches.chapter == 2
    assert matches.recipe_number == 3
    assert matches.recipe_name == "test recipe"
    assert matches.count == 1
    assert matches.__repr__() == (
        "SearchMatch(chapter=2, number=3, line_number=1, count=1)"
    )



"""
You have code that uses a while loop to iteratively process data because it
involves a function or some kind of unusual test condition that doesn't fall
into the usual iteration pattern.
"""
import sys
from pathlib import Path

CHUNKSIZE = 1024


def main():
    """
    ######################################################
    # Common pattern
    def process_data(data):
        pass


    def reader_orig(s):
        while True:
            data = s.recv(CHUNKSIZE)
            if data == b'':
                break
            process_data(data)


    ######################################################
    # Better approach - iter takes optional sentinel param
    def reader(s):
        for chunk in iter(lambda: s.recv(CHUNKSIZE, b'')):
            process_data(chunk)
    """
    with open(Path(__file__).parent / "access-log", "r") as f:
        for chunk in iter(lambda: f.read(10), ""):
            sys.stdout.write(chunk)


if __name__ == "__main__":
    main()



"""
You want to strip unwanted characters, such as whitespace, from the
beginning, end or middle of a text string.
"""


def main():
    s = "     hello world     \n"
    print("s:", s)
    print("s.strip(): ", s.strip())
    print("s.lstrip():", s.lstrip())
    print("s.rstrip():", s.rstrip())
    print()

    t = "-----hello====="
    print("t:", t)
    print("t.lstrip('-'):", t.lstrip("-"))
    print("t.rstrip('='):", t.rstrip("="))
    print("t.strip('-='):", t.strip("-="))


if __name__ == "__main__":
    main()



"""
You want to create a pool of worker threads for serving clients or
performing other kinds of work.
"""

from socket import AF_INET, SOCK_STREAM, socket
from concurrent.futures import ThreadPoolExecutor


def echo_client(sock, client_addr):
    """Handle a client connection"""
    print(f"Got connection from {client_addr}")
    while True:
        msg = sock.recv(65536)
        if not msg:
            break
        sock.sendall(msg)
    print("Client closed connection")
    sock.close()


def echo_server(addr):
    pool = ThreadPoolExecutor(128)
    sock = socket(AF_INET, SOCK_STREAM)
    sock.bind(addr)
    sock.listen(5)
    while True:
        client_sock, client_addr = sock.accept()
        pool.submit(echo_client, client_sock, client_addr)


def main():
    # echo_server(('', 15000))
    pass


if __name__ == "__main__":
    main()



"""
You have a list of dictionaries and you would like to sort the
entries according to one or more of the dictionary values.
"""

from operator import itemgetter
from pprint import pprint


def main():
    rows = [
        {"fname": "Brian", "lname": "Jones", "uid": 1003},
        {"fname": "David", "lname": "Beazley", "uid": 1002},
        {"fname": "John", "lname": "Cleese", "uid": 1001},
        {"fname": "Big", "lname": "Jones", "uid": 1004},
    ]

    print("Rows:")
    pprint(rows)

    print("\nSorted by fname")
    pprint(sorted(rows, key=itemgetter("fname")))

    print("\nSorted by lname")
    pprint(sorted(rows, key=itemgetter("lname")))

    print("\nSorted by both lname & fname")
    pprint(sorted(rows, key=itemgetter("lname", "fname")))

    print("\nRecord with min id")
    pprint(min(rows, key=itemgetter("uid")))

    print("\nRecord with max uid")
    pprint(max(rows, key=itemgetter("uid")))


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from pathlib import Path
from importlib import import_module
import re
from .utils.text import clean_text, extract_leading_numbers


@dataclass
class SearchMatch:
    line_number: int
    line_text: str
    chapter: int
    recipe_number: int
    recipe_name: str
    count: int

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"chapter={self.chapter}, "
            f"number={self.recipe_number}, "
            f"line_number={self.line_number}, "
            f"count={self.count})"
        )


class Recipe:
    module = "example"

    def __init__(self, recipe_dir: Path):
        self.recipe_dir = recipe_dir
        self.path = recipe_dir / f"{self.module}.py"

    @property
    def package(self):
        if self.exists():
            return f"pyrecipes.{self.path.parent.parent.parent.stem}"

    @property
    def chapter_name(self):
        return self.recipe_dir.parent.stem

    @property
    def name(self):
        return self.recipe_dir.stem

    @property
    def clean_name(self):
        return clean_text(self.name)

    @property
    def number(self):
        return extract_leading_numbers(self.name)

    @property
    def chapter(self):
        return extract_leading_numbers(self.chapter_name)

    def exists(self):
        return self.path.exists()

    def get_module(self):
        if not self.exists():
            raise ModuleNotFoundError(f"This recipe couldn't be found:\n  {self.path}")
        return import_module(
            f"{self.package}.{self.chapter_name}.{self.name}.{self.module}"
        )

    def get_docstring(self):
        return self.get_module().__doc__.strip()

    def get_code(self):
        return self.path.read_text()

    def run(self):
        if self.exists():
            print(f"Running {self.chapter}.{self.number} \n")
            getattr(self.get_module(), "main")()
            print()
        else:
            print(f"Couldn't find Recipe {self.name}")

    def search(self, pattern, ignore_case: bool = False):
        results = []
        flags = re.IGNORECASE if ignore_case else 0

        with self.path.open() as file:
            for i, line in enumerate(file, start=1):
                matches = re.findall(re.compile(pattern, flags=flags), line)
                if matches:
                    results.append(
                        SearchMatch(
                            line_number=i,
                            line_text=line,
                            chapter=self.chapter,
                            recipe_number=self.number,
                            recipe_name=self.clean_name,
                            count=len(matches),
                        )
                    )
        return results

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(chapter={self.chapter}, number={self.number})"
        )

    def __str__(self) -> str:
        return f"{self.chapter}.{self.clean_name}"



"""
You need to create or test for the floating-point values of infinity,
negative infinity or NaN (not a number).
"""
import math


def main():
    a = float("inf")
    b = float("-inf")
    c = float("nan")

    print("a:", a, "b:", b, "c:", c)
    print(math.isinf(a))
    print(math.isinf(b))
    print(math.isnan(a))
    print(math.isnan(c))

    print(10 * a)
    print(100 / c)

    # Subtlety - NaNs never compare as equal
    d, e = float("nan"), float("nan")
    print(d == e)


if __name__ == "__main__":
    main()



"""
You want to create a string in which embedded variable names are substituted
with a string representation of a variables value.
"""
import sys


class Info:
    def __init__(self, name, n):
        self.name = name
        self.n = n

    def __str__(self) -> str:
        return f"Info(name={self.name}, n={self.n})"


class safesub(dict):
    def __missing__(self, key):
        return "{" + key + "}"


def sub(text: str):
    return text.format_map(safesub(sys._getframe(1).f_locals))


def main():
    s = "{name} has {n} messages."
    print(s)
    print(s.format(name="Guido", n=37))

    name = "Guido"
    n = 37
    print(s.format_map(vars()))

    a = Info("Guido", 37)
    print(a)
    print(s.format_map(vars(a)))

    try:
        s.format(name="Guido")
    except KeyError as exc:
        print(exc.__class__, exc)

    del n

    print(s.format_map(safesub(vars())))

    name = "Guido"
    n = 37

    print(sub(f"Hello {name}"))
    print(sub(f"You have {n}"))
    print(sub("Your favourite color is {color}"))


if __name__ == "__main__":
    main()



"""
You're writing a class, but you want users to be able to create
instances in more than the one way provided by __init__().
"""
import time


class Date:
    # Primary constructor
    def __init__(self, year, month, day):
        self.year = year
        self.month = month
        self.day = day

    # Alternate constructor
    @classmethod
    def today(cls):
        t = time.localtime()
        return cls(t.tm_year, t.tm_mon, t.tm_mday)


def main():
    a = Date(2012, 12, 21)
    b = Date.today()
    print(a.year, a.month, a.day)
    print(b.year, b.month, b.day)

    class NewDate(Date):
        pass

    c = Date.today()
    d = NewDate.today()
    print(c)
    print(d)
    print("Should be Date instance:", isinstance(c, Date))
    print("Should be NewDate instance:", isinstance(d, NewDate))


if __name__ == "__main__":
    main()



import pytest
from pyrecipes.chapter import Chapter
from pyrecipes.cookbook import CookBook
from pyrecipes.errors import ChapterNotFoundError


@pytest.fixture
def cookbook(recipe_root_dir):
    yield CookBook(recipe_root_dir)


def test_CookBook_init(cookbook, recipe_root_dir):
    assert cookbook.cookbook_dir == recipe_root_dir
    assert cookbook.chapters.keys() == {1, 2, 3}
    assert cookbook.size == 9


def test_CookBook_indexing__exists(cookbook):
    assert isinstance(cookbook[1], Chapter)


def test_CookBook_indexing__doesnt_exist(cookbook):
    with pytest.raises(ChapterNotFoundError, match="Chapter 100 not found"):
        _ = cookbook[100]


def test_iterating_over_CookBook(cookbook):
    for chapter, expected in zip(cookbook, [1, 2, 3]):
        assert chapter.number == expected
        assert isinstance(chapter, Chapter)



from pathlib import Path
from pyrecipes import COOKBOOK_DIR
from pyrecipes.chapter import Chapter
from pyrecipes.errors import ChapterNotFoundError


class CookBook:
    def __init__(self, cookbook_dir: Path = COOKBOOK_DIR) -> None:
        self.cookbook_dir = cookbook_dir
        self.chapters: dict[int, Chapter] = {}
        self._collect()

    def _collect(self):
        """Collects all chapters in cookbok_dir"""
        for chapter_dir in sorted(self.cookbook_dir.glob("[0-9]*")):
            chapter = Chapter(chapter_dir)
            self.chapters[chapter.number] = chapter

    @property
    def size(self):
        return sum(chapter.size for chapter in self)

    def search(self, pattern: str, ignore_case: bool = False):
        results = {}
        for chapter in self.chapters.values():
            matches = chapter.search(pattern, ignore_case)
            if matches:
                results.setdefault(chapter, {}).update(matches)
        return results

    def __getitem__(self, key):
        chapter = self.chapters.get(key)
        if chapter:
            return self.chapters.get(key)
        raise ChapterNotFoundError(key)

    def __iter__(self):
        for chapter in self.chapters.values():
            yield chapter


cookbook = CookBook(COOKBOOK_DIR)



"""
You want a function to only accept certain arguments by keyword.
"""


def recv(maxsize, *, block):
    """Receives message"""
    print(f"maxsize: {maxsize}, block: {block}")


def minimum(*values, clip=None):
    m = min(values)
    if clip is not None:
        return max(clip, m)
    return m


def example_1():
    try:
        # This fails - TtypeError
        recv(1024, True)
    except TypeError:
        recv(1024, block=True)
    print("=" * 20)


def example_2():
    a = [1, 5, 2, -5, 10]

    print("a:", a)
    print("minimum (no clip):", minimum(*a))
    print("minimum (clip=0):", minimum(*a, clip=0))


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
Instead of iterating over a file by lines, you want to iterate
over a collection of fixed-sized records or chunks.
"""
from pathlib import Path
from functools import partial

RECORD_SIZE = 32


def main():
    with (Path(__file__).parent / "somefile.data").open("rb") as f:
        records = iter(partial(f.read, RECORD_SIZE), b"")
        for record in records:
            print(record)


if __name__ == "__main__":
    main()



"""
You have a program that performs a lot of CPU-intensive work
and you would like to make it run faster by having it take
advantage of multiple CPUs.
"""

import gzip
import io
import time
from pathlib import Path
from concurrent import futures


def find_robots(filename):
    """Find all of the hosts that access robots.txt in a single log file."""
    robots = set()
    with gzip.open(filename) as f:
        for line in io.TextIOWrapper(f, encoding="ascii"):
            fields = line.split()
            if fields[6] == "/robots.txt":
                robots.add(fields[0])
    return robots


def find_all_robots(logdir):
    """Find all hosts across the entire sequence of files."""
    files = Path(logdir).glob("./*log.gz")
    all_robots = set()
    for robots in map(find_robots, files):
        all_robots.update(robots)
    return all_robots


def find_all_robots_parallel(logdir):
    """Find all hosts across the entire sequence of files."""
    files = Path(logdir).glob("./*log.gz")
    all_robots = set()
    with futures.ProcessPoolExecutor() as pool:
        for robots in pool.map(find_robots, files):
            all_robots.update(robots)
    return all_robots


def example_1():
    start = time.time()
    robots = find_all_robots(Path(__file__).parent / "logs")
    for ipaddr in robots:
        print(ipaddr)
    print(f"Took {time.time() - start}")


def example_2():
    start = time.time()
    robots = find_all_robots_parallel(Path(__file__).parent / "logs")
    for ipaddr in robots:
        print(ipaddr)
    print(f"Took {time.time() - start}")


def main():
    example_1()
    print()
    example_2()


if __name__ == "__main__":
    main()



import click
from pyrecipes.cookbook import cookbook
from pyrecipes.errors import ChapterNotFoundError, RecipeNotFoundError


@click.command()
@click.argument("chapter", type=int)
@click.argument("number", type=int)
def run(chapter, number):
    """Runs a recipe"""
    try:
        cookbook[chapter][number].run()
    except (ChapterNotFoundError, RecipeNotFoundError) as exc:
        click.echo(exc)



"""
You want to write a function that accepts any number of input arguments.
"""

import html


def avg(first, *rest):
    return (first + sum(rest)) / (1 + len(rest))


def make_element(name, value, **attrs):
    attr_string = "".join(f' {k}="{v}"' for k, v in attrs.items())
    return f"<{name}{attr_string}>{html.escape(value)}</{name}>"


def anyargs(*args, **kwargs):
    return f"args: {args}\nkwargs: {kwargs}"


def main():
    print(f"avg(1, 2): {avg(1, 2)}")
    print(f"avg(1, 2, 3, 4): {avg(1, 2, 3, 4)}")

    print(make_element("item", "Albatross", size="large", quantity=6))
    print(make_element("p", "<spam>"))

    print(anyargs(1, 2, 3, a=4, b=5, c=6))


if __name__ == "__main__":
    main()



"""
Your program has become and unreadable mess of hardcoded slice indices and you want to clean it up.
"""

#         0123456789012345678901234567890123456789012345678901234567890
record = "....................100          .......513.25      ........."


def example_1():
    print(f'\n{"=" * 50}\nExample 1\n{"=" * 50}')
    print(f"Record: {record}\n")
    print(
        "parsing the record with the hardcoded slices [20:32] and [40:48] to find the product..."
    )
    qty = int(record[20:32])
    value = float(record[40:48])
    print(f"Qty: {qty}")
    print(f"Value: {value}")
    print(f"Total: {qty * value}")


def example_2():
    print(f'\n{"=" * 50}\nExample 2\n{"=" * 50}')
    print(f"Record: {record}\n")
    print("parsing the record with the named slices to find the product...\n")
    qty = slice(20, 32)
    value = slice(40, 48)
    total = int(record[qty]) * float(record[value])

    print(f"Qty: {qty}")
    print(f"Value: {value}")
    print(f"Total: {total}")


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You want to perform various calculations (e.g. minimum value,
maximum value, sorting etc.) on a dictionary pf data.
"""


def main():
    prices = {"ACME": 45.23, "AAPL": 612.78, "IBM": 205.55, "HPQ": 37.20, "FB": 10.75}
    print("prices:", prices)

    min_price = min(zip(prices.values(), prices))
    print("min_price:", min_price)

    max_price = max(zip(prices.values(), prices))
    print("max_price:", max_price)

    prices_sorted = sorted(zip(prices.values(), prices))
    print("prices_sorted:", prices_sorted)


if __name__ == "__main__":
    main()



"""
You want to implement a state machine or an object that operates in
a number of different states, but don't want to litter you code with
a lot of conditionals.
"""


def example_1():
    class Connection:
        def __init__(self):
            self.new_state(ClosedConnection)

        def new_state(self, state):
            self.__class__ = state

        def read(self):
            raise NotImplementedError()

        def write(self, data):
            raise NotImplementedError()

        def open(self):
            raise NotImplementedError()

        def close(self):
            raise NotImplementedError()

    class ClosedConnection(Connection):
        def read(self):
            raise RuntimeError("Not open")

        def write(self, data):
            raise RuntimeError("Not open")

        def open(self):
            self.new_state(OpenConnection)

        def close(self):
            raise RuntimeError("Already closed")

    class OpenConnection(Connection):
        def read(self):
            print("reading")

        def write(self, data):
            print("writing")

        def open(self):
            raise RuntimeError("Already open")

        def close(self):
            self.new_state(ClosedConnection)

    c = Connection()
    print(c)
    try:
        c.read()
    except RuntimeError as e:
        print(e)

    c.open()
    print(c)
    c.read()
    c.close()
    print(c)
    print()


def example_2():
    class Connection:
        def __init__(self):
            self.new_state(ClosedConnectionState)

        def new_state(self, newstate):
            self._state = newstate

        # Delegate to the state class
        def read(self):
            return self._state.read(self)

        def write(self, data):
            return self._state.write(self, data)

        def open(self):
            return self._state.open(self)

        def close(self):
            return self._state.close(self)

    # Connection state base class
    class ConnectionState:
        @staticmethod
        def read(conn):
            raise NotImplementedError()

        @staticmethod
        def write(conn, data):
            raise NotImplementedError()

        @staticmethod
        def open(conn):
            raise NotImplementedError()

        @staticmethod
        def close(conn):
            raise NotImplementedError()

    # Implementation of different states
    class ClosedConnectionState(ConnectionState):
        @staticmethod
        def read(conn):
            raise RuntimeError("Not open")

        @staticmethod
        def write(conn, data):
            raise RuntimeError("Not open")

        @staticmethod
        def open(conn):
            conn.new_state(OpenConnectionState)

        @staticmethod
        def close(conn):
            raise RuntimeError("Already closed")

    class OpenConnectionState(ConnectionState):
        @staticmethod
        def read(conn):
            print("reading")

        @staticmethod
        def write(conn, data):
            print("writing")

        @staticmethod
        def open(conn):
            raise RuntimeError("Already open")

        @staticmethod
        def close(conn):
            conn.new_state(ClosedConnectionState)

    c = Connection()
    print(c)
    try:
        c.read()
    except RuntimeError as e:
        print(e)

    c.open()
    print(c)
    c.read()
    c.close()
    print(c)
    print()


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You want to match or search text for a specific pattern.
"""

import re


def main():
    # Some sample text
    text = "Today is 11/27/2012. PyCon starts 3/13/2013."

    # (a) Find all matching dates
    datepat = re.compile(r"\d+/\d+/\d+")
    print(datepat.findall(text))

    # (b) Find all matching dates with capture groups
    datepat = re.compile(r"(\d+)/(\d+)/(\d+)")
    for month, day, year in datepat.findall(text):
        print("{}-{}-{}".format(year, month, day))

    # (c) Iterative search
    for m in datepat.finditer(text):
        print(m.groups())


if __name__ == "__main__":
    main()



"""
You want to round a floating-point number to a fixed numer or decimal place
"""


def example_1():
    x = 1.234567
    print(f"x:\n  {x}\n")
    print(f"rounded 1 decimal place:\n  {round(x, 1)}\n")
    print(f"rounded 3 decimal place:\n  {round(x, 3)}\n")


def example_2():
    x = 1627731
    print(f"x:\n {x}\n")
    print(f"rounded to nearest 10s:\n  {round(x, -1)}\n")
    print(f"rounded to nearest 100s:\n  {round(x, -2)}\n")
    print(f"rounded to nearest 1000s:\n  {round(x, -3)}\n")


def main():
    example_1()
    print("=" * 50)
    example_2()


if __name__ == "__main__":
    main()



"""
You want to take a slice of data produced by an iterator, but the normal
slicing operator doesn't work.
"""

from itertools import islice


def count(n):
    while True:
        yield n
        n += 1


def main():
    print("Printing a slice of a generator function, 10-20")
    for x in islice(count(0), 10, 20):
        print(x)

    print("list slice from generator function")
    print(list(islice(count(0), 4, 12)))


if __name__ == "__main__":
    main()



"""
You want to write a decorator that takes arguments.
"""
from functools import wraps
import logging


def logged(level, name=None, message=None):
    """
    Add logging to a function. level is the logging level,
    name is the logger name and message is the log message.
    If name and message aren't specified, they default to
    the function's module and name.
    """

    def decorate(func):
        logname = name if name else func.__module__
        log = logging.getLogger(logname)
        logmsg = message if message else func.__name__

        @wraps(func)
        def wrapper(*args, **kwargs):
            log.log(level, logmsg)
            return func(*args, **kwargs)

        return wrapper

    return decorate


@logged(logging.INFO)
def add(a, b):
    return a + b


@logged(logging.CRITICAL, message="example message")
def spam():
    print("this is spam")


def main():
    print("Example 1: @logged(logging.INFO)")
    add(1, 2)

    print(
        "\nExample 2: @logged(logging.CRITICAL, name='SPAM!', message='example message')"
    )
    spam()


if __name__ == "__main__":
    main()



import pytest
from pyrecipes import COOKBOOK_DIR, TEMPLATE_PATH
import pyrecipes
from pyrecipes.recipe import Recipe


TEMPLATE = TEMPLATE_PATH.read_text()


@pytest.fixture(params=list(COOKBOOK_DIR.glob("**/example.py")))
def recipe_path(request):
    """Iters over every recipe."""
    yield request.param.parent


@pytest.fixture(scope="session")
def recipe_root_dir(tmp_path_factory):
    """Creates dummy recipe dir with chapters 1, 2, 3 each with 3 example recipes."""
    root_dir = tmp_path_factory.mktemp("recipes")
    (root_dir / "template.py").write_text(TEMPLATE)

    for i in range(1, 4):
        chapter_dir = root_dir / f"0{i}_test_chapter"
        for j in range(1, 4):
            recipe_dir = chapter_dir / f"0{j}_test_recipe"
            recipe_dir.mkdir(parents=True, exist_ok=True)
            (recipe_dir / "example.py").write_text(TEMPLATE)
    yield root_dir


@pytest.fixture
def valid_recipe(recipe_root_dir):
    yield Recipe(recipe_root_dir / "01_test_chapter" / "01_test_recipe")


@pytest.fixture
def patched_root(recipe_root_dir, monkeypatch):
    """Patches the real root recipe directory for the testing dummy one"""
    monkeypatch.setattr(pyrecipes, "ROOT", recipe_root_dir)
    yield monkeypatch



"""
You want to make your objects support the context-management
protocol (the with statement).
"""
from socket import socket, AF_INET, SOCK_STREAM
from functools import partial


class LazyConnection:
    def __init__(self, address, family=AF_INET, type=SOCK_STREAM):
        self.address = address
        self.family = family
        self.type = type
        self.sock = None

    def __enter__(self):
        if self.sock is not None:
            raise RuntimeError("Already connected")
        self.sock = socket(self.family, self.type)
        self.sock.connect(self.address)
        return self.sock

    def __exit__(self, exc_ty, exc_val, tb):
        self.sock.close()
        self.sock = None


def main():
    conn = LazyConnection(("www.python.org", 80))
    # Connection closed
    with conn as s:
        print("conn.__enter__() executes: connection opens", end="\n\n")
        s.send(b"GET /index.html HTTP/1.0\r\n")
        s.send(b"Host: www.python.org\r\n")
        s.send(b"\r\n")
        resp = b"".join(iter(partial(s.recv, 8192), b""))
        print(resp.decode())
        print("conn.__exit__() executes: connection closes")


if __name__ == "__main__":
    main()



"""
You want to eliminate the duplicate values in a sequence,
but preserve the order of the remaining items
"""


def dedupe_hashable(items):
    seen = set()
    for item in items:
        if item not in seen:
            yield item
            seen.add(item)


def dedupe_unhashable(items, key):
    seen = set()
    for item in items:
        val = item if key is None else key(item)
        if val not in seen:
            yield item
            seen.add(val)


def main():
    a = [1, 5, 2, 1, 9, 1, 5, 10]
    b = [{"x": 1, "y": 2}, {"x": 1, "y": 3}, {"x": 1, "y": 2}, {"x": 2, "y": 4}]

    print(f"\nDe-duping {a} (hashable)")
    print(list(dedupe_hashable(a)))
    print("=" * 50)

    print(f"\nDe-duping {b} (unhashable)")
    print(" (de-duping on the x and y key values)")
    print(list(dedupe_unhashable(b, key=lambda d: (d["x"], d["y"]))))
    print("=" * 50)

    print(f"\nDe-duping {b} (unhashable)")
    print(" (de-duping on the x key value)")
    print(list(dedupe_unhashable(b, key=lambda d: d["x"])), end="\n\n")


if __name__ == "__main__":
    main()



"""
You want to optionally enforce type checking of function arguments
as a kind of assertion or contract.
"""

from inspect import signature
from functools import wraps


def typeassert(*ty_args, **ty_kwargs):
    def decorate(func):
        # If in optimized mode disable type-checking
        if not __debug__:
            return func

        # Map function argument names to supplied types
        sig = signature(func)
        bound_types = sig.bind_partial(*ty_args, **ty_kwargs).arguments

        @wraps(func)
        def wrapper(*args, **kwargs):
            bound_values = sig.bind(*args, **kwargs)

            # Enforce type assertions across supplied arguments
            for name, value in bound_values.arguments.items():
                if name in bound_types:
                    if not isinstance(value, bound_types[name]):
                        raise TypeError(f"Argument {name} must be {bound_types[name]}")
            return func(*args, **kwargs)

        return wrapper

    return decorate


@typeassert(int, int)
def add(a, b):
    return a + b


@typeassert(int, z=str)
def spam(x, y, z):
    return f"{'SPAM! ' * x} - y: {y} - filed by {z}"


def main():
    print(add(1, 2))

    try:
        print(add(1, "a"))
    except TypeError as exc:
        print(f"Caught error: {exc}")

    print(spam(3, 2, "Chris"))

    try:
        print(spam(3, 2, 9.0))
    except TypeError as exc:
        print(f"Caught error: {exc}")


if __name__ == "__main__":
    main()



"""
You've learned about function argument annotations and you have thought
that you might be able to use them to implement multiple-dispatch
(method overloading) based on types. However, you're not quite sure
what's involved (or if it's even a good idea).
"""

import inspect
import types


def example_1():
    class MultiMethod:
        """
        Represents a single multimethod.
        """

        def __init__(self, name):
            self._methods = {}
            self.__name__ = name

        def register(self, meth):
            """
            Register a new method as a multimethod
            """
            sig = inspect.signature(meth)

            # Build a type-signature from the method's annotations
            types = []
            for name, parm in sig.parameters.items():
                if name == "self":
                    continue
                if parm.annotation is inspect.Parameter.empty:
                    raise TypeError(
                        "Argument {} must be annotated with a type".format(name)
                    )
                if not isinstance(parm.annotation, type):
                    raise TypeError(
                        "Argument {} annotation must be a type".format(name)
                    )
                if parm.default is not inspect.Parameter.empty:
                    self._methods[tuple(types)] = meth
                types.append(parm.annotation)

            self._methods[tuple(types)] = meth

        def __call__(self, *args):
            """
            Call a method based on type signature of the arguments
            """
            types = tuple(type(arg) for arg in args[1:])
            meth = self._methods.get(types, None)
            if meth:
                return meth(*args)
            else:
                raise TypeError("No matching method for types {}".format(types))

        def __get__(self, instance, cls):
            """
            Descriptor method needed to make calls work in a class
            """
            if instance is not None:
                return types.MethodType(self, instance)
            else:
                return self

    class MultiDict(dict):
        """
        Special dictionary to build multimethods in a metaclass
        """

        def __setitem__(self, key, value):
            if key in self:
                # If key already exists, it must be a multimethod or callable
                current_value = self[key]
                if isinstance(current_value, MultiMethod):
                    current_value.register(value)
                else:
                    mvalue = MultiMethod(key)
                    mvalue.register(current_value)
                    mvalue.register(value)
                    super().__setitem__(key, mvalue)
            else:
                super().__setitem__(key, value)

    class MultipleMeta(type):
        """
        Metaclass that allows multiple dispatch of methods
        """

        def __new__(cls, clsname, bases, clsdict):
            return type.__new__(cls, clsname, bases, dict(clsdict))

        @classmethod
        def __prepare__(cls, clsname, bases):
            return MultiDict()

    # Some example classes that use multiple dispatch
    class Spam(metaclass=MultipleMeta):
        def bar(self, x: int, y: int):
            print("Bar 1:", x, y)

        def bar(self, s: str, n: int = 0):
            print("Bar 2:", s, n)

    # Example: overloaded __init__
    import time

    class Date(metaclass=MultipleMeta):
        def __init__(self, year: int, month: int, day: int):
            self.year = year
            self.month = month
            self.day = day

        def __init__(self):
            t = time.localtime()
            self.__init__(t.tm_year, t.tm_mon, t.tm_mday)

    s = Spam()
    s.bar(2, 3)
    s.bar("hello")
    s.bar("hello", 5)
    try:
        s.bar(2, "hello")
    except TypeError as e:
        print(e)

    # Overloaded __init__
    d = Date(2012, 12, 21)
    print(d.year, d.month, d.day)
    # Get today's date
    e = Date()
    print(e.year, e.month, e.day)


def example_2():
    class multimethod:
        def __init__(self, func):
            self._methods = {}
            self.__name__ = func.__name__
            self._default = func

        def match(self, *types):
            def register(func):
                ndefaults = len(func.__defaults__) if func.__defaults__ else 0
                for n in range(ndefaults + 1):
                    self._methods[types[: len(types) - n]] = func
                return self

            return register

        def __call__(self, *args):
            types = tuple(type(arg) for arg in args[1:])
            meth = self._methods.get(types, None)
            if meth:
                return meth(*args)
            else:
                return self._default(*args)

        def __get__(self, instance, cls):
            if instance is not None:
                return types.MethodType(self, instance)
            else:
                return self

    # Example use
    class Spam:
        @multimethod
        def bar(self, *args):
            # Default method called if no match
            raise TypeError("No matching method for bar")

        @bar.match(int, int)
        def bar(self, x, y):
            print("Bar 1:", x, y)

        @bar.match(str, int)
        def bar(self, s, n=0):
            print("Bar 2:", s, n)

    s = Spam()
    s.bar(2, 3)
    s.bar("hello")
    s.bar("hello", 5)
    try:
        s.bar(2, "hello")
    except TypeError as e:
        print(e)


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You have multiple dictionaries or mappings that you want to logically
combine into a single mapping to perform certain operations, such as
looking up values or checking for the existence of keys.
"""

from collections import ChainMap


def main():
    a = {"x": 1, "z": 3}
    b = {"y": 2, "z": 4}

    print("dict a:", a)
    print("dict b:", a)

    c = ChainMap(a, b)
    print("ChainMap c:", c)
    for x in list("xyz"):
        print(f"ChainMap[{x}]:", c[x])
    print("ChainMap keys:", list(c.keys()))
    print("ChainMap values:", list(c.values()))
    print("ChainMap items:", list(c.items()))


if __name__ == "__main__":
    main()



"""
You want to perform I/O operations using raw filenames that have not
been decoded or encoded according to the default filename encoding.
"""
import sys
import os
from pathlib import Path

parent_dir = Path(__file__).parent
parent_dir_encoded = str(parent_dir).encode()


def main():
    print(sys.getfilesystemencoding())

    # Write a file using unicode filename
    with parent_dir.joinpath("jalape\xf1o.txt").open("w") as f:
        f.write("Spicy!")

    # Directory listing (decoded)
    txt_files = [file for file in os.listdir(parent_dir) if file.endswith(".txt")]
    print(txt_files)

    # Directory listing (raw)
    txt_files = [
        file for file in os.listdir(parent_dir_encoded) if file.endswith(b".txt")
    ]
    print(txt_files)

    # Opening file with raw name
    with open(os.path.join(parent_dir_encoded, txt_files[0]), "r") as f:
        print(f.read())


if __name__ == "__main__":
    main()



"""
You want to iterate over an iterable, but the first few items aren't of interest
and you just want to discard them.
"""

from itertools import dropwhile
from pathlib import Path


def main():
    with open(Path(__file__).parent / "somefile.txt", "r") as f:
        for line in dropwhile(lambda line: line.startswith("#"), f):
            print(line, end="")


if __name__ == "__main__":
    main()



"""
You would like to define a generator function, but it involves extra
state that you would like to expose to the user somehow.
"""

from collections import deque
from pathlib import Path


class LineHistory:
    def __init__(self, lines, history=3):
        self.lines = lines
        self.history = deque(maxlen=history)

    def __iter__(self):
        for i, line in enumerate(self.lines, 1):
            self.history.append((i, line))
            yield line

    def clear(self):
        self.history.clear()


def main():
    with open(Path(__file__).parent / "somefile.txt", "r") as f:
        lines = LineHistory(f)
        for line in lines:
            if "python" in line:
                print(
                    f'found python in line "{line}" - here is the current history state.'
                )
                for line_num, history_line in lines.history:
                    print(f"{line_num}:{history_line}", end="")


if __name__ == "__main__":
    main()



"""
You want to define a metaclass that allows class definitions to supply
optional arguments, possiblly to control or configure aspects of processing
during type creation.
"""

from abc import ABCMeta, abstractmethod


class IStream(metaclass=ABCMeta):
    @abstractmethod
    def read(self, maxsize=None):
        pass

    @abstractmethod
    def write(self, data):
        pass


class MyMeta(type):
    # Optional
    @classmethod
    def __prepare__(cls, name, bases, *, debug=False, synchronize=False):
        # Custom processing
        ...
        return super().__prepare__(name, bases)

    # Required
    def __new__(cls, name, bases, ns, *, debug=False, synchronize=False):
        # Custom processing
        ...
        super().__new__(cls, name, bases, ns)

    # Required
    def __init__(self, name, bases, ns, *, debug=False, synchronize=False):
        # Custom processing
        super().__init__(name, bases, ns)


class Spam(metaclass=MyMeta):
    debug = True
    synchronize = True
    ...


def main():
    pass


if __name__ == "__main__":
    main()



"""
You have data inside a sequence and need to extract values or reduce a
sequence using some criteria.
"""


def main():
    # Examples of different ways to filter data

    mylist = [1, 4, -5, 10, -7, 2, 3, -1]

    # All positive values
    pos = [n for n in mylist if n > 0]
    print(pos)

    # All negative values
    neg = [n for n in mylist if n < 0]
    print(neg)

    # Negative values clipped to 0
    neg_clip = [n if n > 0 else 0 for n in mylist]
    print(neg_clip)

    # Positive values clipped to 0
    pos_clip = [n if n < 0 else 0 for n in mylist]
    print(pos_clip)

    # Compressing example

    addresses = [
        "5412 N CLARK",
        "5148 N CLARK",
        "5800 E 58TH",
        "2122 N CLARK",
        "5645 N RAVENSWOOD",
        "1060 W ADDISON",
        "4801 N BROADWAY",
        "1039 W GRANVILLE",
    ]

    counts = [0, 3, 10, 4, 1, 7, 6, 1]

    from itertools import compress

    more5 = [n > 5 for n in counts]
    a = list(compress(addresses, more5))
    print(a)


if __name__ == "__main__":
    main()



"""
You have a collection of generally useful methods that you would
like to make available for extending the functionality of other
class definitions. However, the classes where the methods might
be added aren't necessarily related to one another via inheritance.
Thus, you can't just attach the methods to a common base class.
"""


def example_1():
    class LoggedMappingMixin:
        """
        Add logging to get/set/delete operations for debugging.
        """

        __slots__ = ()

        def __getitem__(self, key):
            print("Getting " + str(key))
            return super().__getitem__(key)

        def __setitem__(self, key, value):
            print("Setting {} = {!r}".format(key, value))
            return super().__setitem__(key, value)

        def __delitem__(self, key):
            print("Deleting " + str(key))
            return super().__delitem__(key)

    class SetOnceMappingMixin:
        """
        Only allow a key to be set once.
        """

        __slots__ = ()

        def __setitem__(self, key, value):
            if key in self:
                raise KeyError(str(key) + " already set")
            return super().__setitem__(key, value)

    class StringKeysMappingMixin:
        """
        Restrict keys to strings only
        """

        __slots__ = ()

        def __setitem__(self, key, value):
            if not isinstance(key, str):
                raise TypeError("keys must be strings")
            return super().__setitem__(key, value)

    # Examples
    print("# ---- LoggedDict Example")

    class LoggedDict(LoggedMappingMixin, dict):
        pass

    d = LoggedDict()
    d["x"] = 23
    print(d["x"])
    del d["x"]

    print("# ---- SetOnceDefaultDict Example")

    from collections import defaultdict

    class SetOnceDefaultDict(SetOnceMappingMixin, defaultdict):
        pass

    d = SetOnceDefaultDict(list)
    d["x"].append(2)
    d["y"].append(3)
    d["x"].append(10)
    try:
        d["x"] = 23
    except KeyError as e:
        print(e)

    print("# ---- StringOrderedDict Example")
    from collections import OrderedDict

    class StringOrderedDict(StringKeysMappingMixin, SetOnceMappingMixin, OrderedDict):
        pass

    d = StringOrderedDict()
    d["x"] = 23
    try:
        d[42] = 10
    except TypeError as e:
        print(e)

    try:
        d["x"] = 42
    except KeyError as e:
        print(e)
    print()


def example_2():
    class RestrictKeysMixin:
        def __init__(self, *args, _restrict_key_type, **kwargs):
            self.__restrict_key_type = _restrict_key_type
            super().__init__(*args, **kwargs)

        def __setitem__(self, key, value):
            if not isinstance(key, self.__restrict_key_type):
                raise TypeError("Keys must be " + str(self.__restrict_key_type))
            super().__setitem__(key, value)

    # Example

    class RDict(RestrictKeysMixin, dict):
        pass

    d = RDict(_restrict_key_type=str)
    e = RDict([("name", "Dave"), ("n", 37)], _restrict_key_type=str)
    f = RDict(name="Dave", n=37, _restrict_key_type=str)
    print(d)
    print(e)
    print(f)
    try:
        f[42] = 10
    except TypeError as e:
        print(e)
    print()


def example_3():
    # Class decorator alternative to mixins
    def LoggedMapping(cls):
        cls_getitem = cls.__getitem__
        cls_setitem = cls.__setitem__
        cls_delitem = cls.__delitem__

        def __getitem__(self, key):
            print("Getting %s" % key)
            return cls_getitem(self, key)

        def __setitem__(self, key, value):
            print("Setting %s = %r" % (key, value))
            return cls_setitem(self, key, value)

        def __delitem__(self, key):
            print("Deleting %s" % key)
            return cls_delitem(self, key)

        cls.__getitem__ = __getitem__
        cls.__setitem__ = __setitem__
        cls.__delitem__ = __delitem__
        return cls

    @LoggedMapping
    class LoggedDict(dict):
        pass

    d = LoggedDict()
    d["x"] = 23
    print(d["x"])
    del d["x"]


def main():
    example_1()
    example_2()
    example_3()


if __name__ == "__main__":
    main()



"""
You need to access various services via HTTP as a client. For example,
downloading or interacting with a REST-based API.
"""

from urllib import request, parse
import requests


def example_1():
    """For very simple requests"""
    # Base URL being accessed
    url = "http://httpbin.org/"

    # Dictionary of query parameters (if any)
    params = {"name1": "value1", "name2": "value2"}

    # Encode the query string
    query_string = parse.urlencode(params)

    # Make GET request
    response = request.urlopen(f"{url}get?{query_string}")
    print(response.read().decode("utf8"))

    # Make POST request
    response = request.urlopen(f"{url}post", query_string.encode("ascii"))
    print(response.read().decode("utf8"))

    # With headers
    headers = {"User-agent": "nunya/biznis", "Spam": "eggs"}
    req = request.Request(f"{url}post", query_string.encode("ascii"), headers=headers)
    response = request.urlopen(req)
    print(response.read().decode("utf8"))


def example_2():
    """For anything more complicated than the previous example, use requests."""
    url = "http://httpbin.org/"

    params = {"name1": "value1", "name2": "value2"}

    headers = {"User-agent": "nunya/biznis", "Spam": "eggs"}

    response = requests.post(f"{url}post", data=params, headers=headers)
    print(response)
    print(response.status_code)
    print(response.json())


def example_3():
    response = requests.head("http://python.org/index.html")
    print(response)
    print(response.headers)


def example_4():
    url = "http://httpbin.org/"
    response1 = requests.get(url)
    response2 = requests.get(url, cookies=response1.cookies)
    print(response1)
    print(response2)


def main():
    # example_1()
    # example_2()
    # example_3()
    # example_4()
    pass


if __name__ == "__main__":
    main()



"""
You need to format text with some sort of alignment applied.
"""


def main():
    text = "Hello World"
    print(text)
    print(text.ljust(20))
    print(text.rjust(20))
    print(text.center(20))
    print(text.rjust(20, "="))
    print(text.center(20, "*"))
    print(format(text, ">20"))
    print(format(text, "=>20"))


if __name__ == "__main__":
    main()



"""
You have a conference call scheduled for December 21, 2024 at 9:30am
in Chicago. At what local time did you friend in Bangalore, India
have to show up to attend?
"""
import pytz
from datetime import datetime, timedelta
from pytz import timezone


def main():
    d = datetime(2024, 12, 21, 9, 30)
    central = timezone("US/Central")
    loc_d = central.localize(d)
    print("d:", d)
    print("central:", central)
    print("localized:", loc_d)

    bang_d = loc_d.astimezone(timezone("Asia/Kolkata"))
    print(bang_d)

    later = central.normalize(loc_d + timedelta(minutes=30))
    print(later)
    print(later.astimezone(timezone("Asia/Kolkata")))

    now_utc = datetime.now(tz=pytz.utc)
    print(now_utc)


if __name__ == "__main__":
    main()



"""
You need to crunch through large datasets and generate summaries or
other kinds of statistics.
"""


def example_1():
    """For illustrative purposes only"""
    import pandas as pd

    # Reading csv
    df = pd.read_csv("rats.csv", skipfooter=1)

    # Getting distrinct values fomr a column
    df["Current Activity"].unique()

    # Filtering
    df[df["Current Activity"] == "Dispatch Crew"]

    # Finding top 10 value
    df["ZIP Code"].value_counts().head(10)

    # Group by
    grp = df.groupby("Completion Date")
    sizes = grp.size()
    sizes.sort_values()


def main():
    pass


if __name__ == "__main__":
    main()



"""
You have 2 dictionaries and want to find out what they might have in common (keys, values etc)
"""

from pprint import pprint

d1 = {"x": 1, "y": 2, "z": 3}

d2 = {"w": 10, "x": 11, "y": 2}


def main():
    print("Dict 1:")
    pprint(d1)
    print("\nDict 2:")
    pprint(d2)
    print(f"\nCommon keys:\n  {d1.keys() & d2.keys()}")
    print(f"Keys in dict 1 that are not in dict 2:\n  {d1.keys() - d2.keys()}")
    print(f"(key, value) pairs in common:\n  {d1.items() & d2.items()}")


if __name__ == "__main__":
    main()



"""
You have a class that only defines a single method besides __init__().
However, to simplify your code, you would much rather just have a simple function.
"""


class URLTemplate:
    def __init__(self, template):
        self.template = template

    def open(self, **kwargs):
        return self.template.format_map(kwargs)


def urltemplate(template):
    def opener(**kwargs):
        return template.format_map(kwargs)

    return opener


def main():
    # Example usage of the class method
    yahoo = URLTemplate("http://finance.yahoo.com/d/quotes.csv?s={names}&f={fields}")
    print(yahoo.open(names="IBM,AAPL,FB", fields="sl1c1v"))

    # Replacing the class with a function using closures
    yahoo = urltemplate("http://finance.yahoo.com/d/quotes.csv?s={names}&f={fields}")
    print(yahoo(names="IBM,AAPL,FB", fields="sl1c1v"))


if __name__ == "__main__":
    main()



"""
You want to write a decorator function that wraps a function, but
has user adjustable attibutes that can be used to control the behavior
of the decorator at runtime.
"""

from functools import partial, wraps
import logging


# Utility decorator to attach a function as an attribute of obj
def attach_wrapper(obj, func=None):
    if func is None:
        return partial(attach_wrapper, obj)
    setattr(obj, func.__name__, func)
    return func


def logged(level, name=None, message=None):
    """Add logging to a function. level is the logging
    level, name is the logger name, message is the log
    message. If name and message aren't specified, they
    default to the functions module and name."""

    def decorate(func):
        logname = name if name else func.__module__
        log = logging.getLogger(logname)
        logmsg = message if message else func.__name__

        @wraps(func)
        def wrapper(*args, **kwargs):
            log.log(level, logmsg)
            return func(*args, **kwargs)

        @attach_wrapper(wrapper)
        def set_level(newlevel):
            nonlocal level
            level = newlevel

        @attach_wrapper(wrapper)
        def set_message(newmsg):
            nonlocal logmsg
            logmsg = newmsg

        return wrapper

    return decorate


@logged(logging.DEBUG)
def add(x, y):
    return x + y


@logged(logging.CRITICAL, "example")
def spam():
    print("Spam!")


def main():
    logging.basicConfig(level=logging.DEBUG)
    print(add(1, 2))
    add.set_message("Add was called.")
    print(add(2, 3))
    add.set_level(logging.WARN)
    print(add(3, 4))
    spam()


if __name__ == "__main__":
    main()



"""
You're working with Unicode strings, but you need to make sure
that all of the strings have the same underlying representation.
"""
import unicodedata


def main():
    s1 = "Spicy Jalape\u00f1o"
    s2 = "Spicy Jalapen\u0303o"
    print("s1:", s1)
    print("s2:", s2)
    print("s1 == s2:", s1 == s2)
    print("len s1:", len(s1))
    print("len s2:", len(s2))
    print()

    t1 = unicodedata.normalize("NFC", s1)
    t2 = unicodedata.normalize("NFC", s2)
    print("t1:", t1)
    print("t2:", t2)
    print("t1 == t2:", t1 == t2)
    print("len t1:", len(t1))
    print("len t2:", len(t2))
    print("t1 ascii:", ascii(t1))
    print("t2 ascii:", ascii(t2))
    print()

    t3 = unicodedata.normalize("NFD", s1)
    t4 = unicodedata.normalize("NFD", s2)
    print("t3:", t3)
    print("t4:", t4)
    print("t3 == t4:", t3 == t4)
    print("len t3:", len(t3))
    print("len t4:", len(t4))
    print("t3 ascii:", ascii(t3))
    print("t4 ascii:", ascii(t4))
    print()

    s = "\ufb01"
    print("s:", s)
    print("s - NFD:", unicodedata.normalize("NFD", s))
    print("s - NFKC:", unicodedata.normalize("NFKC", s))
    print("s - NFKD:", unicodedata.normalize("NFKD", s))


if __name__ == "__main__":
    main()



"""
You want precise control over the symbols that are exported from a module
or package when a user uses the 'from module import *' statement.
"""

import textwrap


def main():
    print(
        textwrap.dedent(
            """
            Simply define a variable '__all__' in your module that explicly
            lists the exported names. For examples:

            # somemodule.py

            def spam():
                pass

            def grok():
                pass

            blah = 42

            # Only export 'spam' and 'grok'
            __all__ = ['spam', 'grok']"""
        )
    )


if __name__ == "__main__":
    main()



"""
You need to perform accurate calculations with decimal numbers, and don't
the small errors that naturally occur with floats.
"""

from decimal import Decimal


def example_floats():
    print("floats...")
    a = 4.2
    b = 2.1
    print("a:", a)
    print("b:", b)
    print("a + b:", a + b)
    print("a + b == 6.3:", (a + b) == 6.3)
    print()


def example_decimals():
    print("decimals...")
    a = Decimal("4.2")
    b = Decimal("2.1")
    print("a:", a)
    print("b:", b)
    print("a + b:", a + b)
    print("a + b == 6.3:", (a + b) == Decimal("6.3"))
    print()


def main():
    example_floats()
    example_decimals()


if __name__ == "__main__":
    main()



"""
You have a module that you would like to split into multiple files. However,
you would like to do it without breaking existing code by keeping the
separate files unified as a single logical module.
"""

from pathlib import Path
from pyrecipes.utils.text import tree


def main():
    example_path = Path(__file__).parent / "src"
    print("Consider the following module, 'my_module.py':\n")
    print("-" * 25)
    print((Path(__file__).parent / "my_module.py").read_text())
    print("-" * 25)
    print(
        "\nThis can be split into a package with multiple modules like the following structure:\n"
    )
    for line in tree(example_path):
        print(line)

    print("\n\nThen each file would have the following code:\n")
    for file in sorted(example_path.glob("my_module/*.py")):
        print("-" * 25)
        print(file.read_text())
        print()


if __name__ == "__main__":
    main()



"""
You want to change the way in which instances are created in order to
implement singletons, caching, or other similar features.
"""

import weakref


class NoInstances(type):
    def __call__(self, *args, **kwargs):
        raise TypeError("Can't instantiate directly.")


class Singleton(type):
    def __init__(self, *args, **kwargs):
        self.__instance = None
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        if self.__instance is None:
            self.__instance = super().__call__(*args, **kwargs)
            return self.__instance
        else:
            return self.__instance


class Cached(type):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__cache = weakref.WeakValueDictionary()

    def __call__(self, *args):
        if args in self.__cache:
            return self.__cache[args]
        else:
            obj = super().__call__(*args)
            self.__cache[args] = obj
            return obj


def example_1():
    class Spam(metaclass=NoInstances):
        @staticmethod
        def grok(x):
            print("Spam.grok:", x)

    print("Example 1 - No instances")
    try:
        spam = Spam()
        spam.grok("foo")
    except TypeError as exc:
        print(exc)

    Spam.grok("foo")


def example_2():
    class Spam(metaclass=Singleton):
        def __init__(self):
            print("Creating spam")

    print("Example 2 - Implements Singleton")
    a = Spam()
    b = Spam()
    print("a is b:", a is b)
    c = Spam()
    print("c ia a:", c is a)


def example_3():

    class Spam(metaclass=Cached):
        def __init__(self, name):
            print(f"Creating Spam({name})")
            self.name = name

    print("Example 3 - Cached implementation")
    a = Spam("foo")
    b = Spam("bar")
    c = Spam("foo")  # cached

    print("a is b:", a is b)
    print("a is b:", a is c)


def main():
    example_1()
    example_2()
    example_3()


if __name__ == "__main__":
    main()



"""
You want to wrap functions with a decorator, but the result is going
to be a callable instance. You need your decorator to work both inside
and outside class definitions.
"""

import types
from functools import wraps


class Profiled:
    def __init__(self, func):
        wraps(func)(self)
        self.ncalls = 0

    def __call__(self, *args, **kwargs):
        self.ncalls += 1
        return self.__wrapped__(*args, **kwargs)

    def __get__(self, instance, cls):
        if instance is None:
            return self
        else:
            return types.MethodType(self, instance)


@Profiled
def add(x, y):
    return x + y


class Spam:
    @Profiled
    def bar(self, x):
        print(self, x)


def main():
    print("add 1 + 2:", add(1, 2))
    print("add 3 + 4:", add(3, 4))
    print("add ncalls:", add.ncalls)

    s = Spam()
    s.bar(2)
    s.bar(3)
    s.bar(4)
    print("Spam bar ncalls:", Spam.bar.ncalls)


if __name__ == "__main__":
    main()



from pyrecipes.utils.text import clean_text, extract_leading_numbers, text_border


def test_clean_text():
    text = "01_this_is_a_test"
    assert clean_text(text) == "1) This is a test"


def test_extract_leading_number():
    text = "02_this_is_another_test"
    assert extract_leading_numbers(text) == 2


def test_assert_text_border():
    text = "testing"
    assert text_border(text) == "===========\n= testing =\n==========="
    assert (
        text_border(text, symbol="+", side_symbol="|", padding=3)
        == "+++++++++++++++\n|   testing   |\n+++++++++++++++"
    )



"""
Your program received a directory listing, but when it tried to print
the filenames, it crashed with a UnicodeEncodeError exception and a
cryptic message about "surrogates not allowed".
"""


def bad_filename(filename):
    return repr(filename[1:-1])


def main():
    filename = b"\udce4d.txt"
    try:
        print(filename)
    except UnicodeEncodeError:
        print("caught bad filename")
        print(bad_filename(filename))


if __name__ == "__main__":
    main()



"""
You need to write code that processes or navigates through a complicated
data structure consisting of many different kinds of objects, each of
which needs to be handled in a different way. For example, walking through
a tree structure and performing different actions depending on what kind of
tree nodes are encountered.
"""


# --- The following classes represent nodes in an expression tree
class Node:
    pass


class UnaryOperator(Node):
    def __init__(self, operand):
        self.operand = operand


class BinaryOperator(Node):
    def __init__(self, left, right):
        self.left = left
        self.right = right


class Add(BinaryOperator):
    pass


class Sub(BinaryOperator):
    pass


class Mul(BinaryOperator):
    pass


class Div(BinaryOperator):
    pass


class Negate(UnaryOperator):
    pass


class Number(Node):
    def __init__(self, value):
        self.value = value


# --- The visitor base class
class NodeVisitor:
    def visit(self, node):
        methname = "visit_" + type(node).__name__
        meth = getattr(self, methname, None)
        if meth is None:
            meth = self.generic_visit
        return meth(node)

    def generic_visit(self, node):
        raise RuntimeError("No {} method".format("visit_" + type(node).__name__))


# --- Example 1:  An expression evaluator
class Evaluator(NodeVisitor):
    def visit_Number(self, node):
        return node.value

    def visit_Add(self, node):
        return self.visit(node.left) + self.visit(node.right)

    def visit_Sub(self, node):
        return self.visit(node.left) - self.visit(node.right)

    def visit_Mul(self, node):
        return self.visit(node.left) * self.visit(node.right)

    def visit_Div(self, node):
        return self.visit(node.left) / self.visit(node.right)

    def visit_Negate(self, node):
        return -node.operand


# --- Example 2: Generate stack instructions
class StackCode(NodeVisitor):
    def generate_code(self, node):
        self.instructions = []
        self.visit(node)
        return self.instructions

    def visit_Number(self, node):
        self.instructions.append(("PUSH", node.value))

    def binop(self, node, instruction):
        self.visit(node.left)
        self.visit(node.right)
        self.instructions.append((instruction,))

    def visit_Add(self, node):
        self.binop(node, "ADD")

    def visit_Sub(self, node):
        self.binop(node, "SUB")

    def visit_Mul(self, node):
        self.binop(node, "MUL")

    def visit_Div(self, node):
        self.binop(node, "DIV")

    def unaryop(self, node, instruction):
        self.visit(node.operand)
        self.instructions.append((instruction,))

    def visit_Negate(self, node):
        self.unaryop(node, "NEG")


def main():
    # Representation of 1 + 2 * (3 - 4) / 5
    t1 = Sub(Number(3), Number(4))
    t2 = Mul(Number(2), t1)
    t3 = Div(t2, Number(5))
    t4 = Add(Number(1), t3)

    e = Evaluator()
    print("Should get 0.6 :", e.visit(t4))

    s = StackCode()
    code = s.generate_code(t4)
    for c in code:
        print(c)


if __name__ == "__main__":
    main()



"""
You want to iterate over the items contained in more than one sequence
at a time.
"""
from itertools import zip_longest


def main():
    xpts = [1, 5, 4, 2, 10, 7]
    ypts = [101, 78, 37, 15, 62, 99]

    print("xpts:", xpts, "| ypts:", ypts)
    for x, y in zip(xpts, ypts):
        print("(x, y):", (x, y))

    a = [1, 2, 3]
    b = ["w", "x", "y", "z"]

    print("\na:", a, " | b:", b)
    print("\ndefault zipping behaviour")
    for i in zip(a, b):
        print("(a, b):", i)

    print("\nusing zip_longest")
    for i in zip_longest(a, b):
        print("(a, b):", i)


if __name__ == "__main__":
    main()



"""
You want to combine many small strings together into a larger string.
"""


def main():
    parts = ["Is", "Chicago", "Not", "Chicago?"]
    print(parts)
    for sep in [" ", ",", ""]:
        print(sep.join(parts))

    a = "Is Chicago"
    b = "Not Chicago?"
    print("a:", a)
    print("b:", b)
    print(a + " " + b)

    a = "hello" "world"
    print(a)


if __name__ == "__main__":
    main()



"""
You're writing code that relies on the use of callback functions
(e.g. even handlers, completion callbacks, etc.), but you want to
have the callback function carry extra state for use inside the
callback function.
"""


def apply_async(func, args, *, callback):
    # Compute the result
    result = func(*args)

    # Invoke the callback with the result
    callback(result)


def print_result(result):
    print("Got:", result)


def add(x, y):
    return x + y


class ResultHandler:
    def __init__(self):
        self.sequence = 0

    def handler(self, result):
        self.sequence += 1
        print(f"[{self.sequence}] Got: {result}")


def example_1():
    """Callback only accepts a single result - no other info is passed."""
    print(example_1.__doc__)
    apply_async(add, (1, 2), callback=print_result)
    apply_async(add, ("Hello ", "World!"), callback=print_result)
    print()


def examlpe_2():
    """Callback is a bound method now - additional state can be passed."""
    r = ResultHandler()
    print(examlpe_2.__doc__)
    apply_async(add, (1, 2), callback=r.handler)
    apply_async(add, ("Hello ", "World!"), callback=r.handler)
    print()


def main():
    example_1()
    examlpe_2()


if __name__ == "__main__":
    main()



"""
You want to inspect or rewrite portions of a class definition
to alter its behavior, but without using inheritance or metaclasses.
"""


def log_getattribute(cls):
    # Get the original implementation
    orig_getattribute = cls.__getattribute__

    # Make a new definition
    def new_getattribute(self, name):
        print("getting:", name)
        return orig_getattribute(self, name)

    # Attach to the class and return
    cls.__getattribute__ = new_getattribute
    return cls


# Example use
@log_getattribute
class A:
    def __init__(self, x):
        self.x = x

    def spam(self):
        pass


def main():
    a = A(42)
    print(a.x)
    a.spam()


if __name__ == "__main__":
    main()



"""
You want to read an XML document, make changes to it, and then
write it back out as XML
"""
import sys
from xml.etree.ElementTree import parse, Element
from pathlib import Path


def main():
    with open(Path(__file__).parent / "pred.xml", "r") as f:
        doc = parse(f)
        print(doc)
    root = doc.getroot()
    print(root)

    # Remove a few elements
    root.remove(root.find("sri"))
    root.remove(root.find("cr"))

    # Insert a new element after <nm>...</nm>
    idx = list(root).index(root.find("nm"))
    print(idx)

    e = Element("spam")
    e.text = "This is a test"
    root.insert(idx + 1, e)

    # Write to stdout
    doc.write(sys.stdout.buffer, xml_declaration=True)


if __name__ == "__main__":
    main()



"""
You want to change the output produced by printing or viewing instances
to something more sensible.
"""


class Pair:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __repr__(self):
        return "Pair({0.x!r}, {0.y!r})".format(self)

    def __str__(self):
        return "({0.x!s}, {0.y!s})".format(self)


def main():
    p = Pair(3, 4)
    print(p)
    print(p.__repr__())


if __name__ == "__main__":
    main()



"""
You have an integer file descriptor corresponding to an already open I/O
channel on the operating system (e.g. file, pipe, socket etc.) and you want
to wrap a higher level Python file object around it.
"""
import os
from socket import socket, AF_INET, SOCK_STREAM
from pathlib import Path

filename = Path(__file__).parent / "somefile.txt"


def echo_client(client_sock, addr):
    print("Got connection from", addr)

    # Make text-mode file wrappers for socket reading/writing
    client_in = open(client_sock.fileno(), "rt", encoding="latin-1", closefd=False)
    client_out = open(client_sock.fileno(), "wt", encoding="latin-1", closefd=False)

    # Echo lines back to the client using file I/O
    for line in client_in:
        client_out.write(line)
        client_out.flush()
    client_sock.close()


def echo_server(address):
    sock = socket(AF_INET, SOCK_STREAM)
    sock.bind(address)
    sock.listen(1)
    while True:
        client, addr = sock.accept()
        echo_client(client, addr)


def example_1():
    # Open a low-level file descriptor
    fd = open(str(filename.resolve()), os.O_WRONLY | os.O_CREAT)

    # Turn into a proper file
    f = open(fd, "wt")
    f.write("hello world\n")
    f.close()


def main():
    pass


if __name__ == "__main__":
    main()



"""
You've written a function or method that uses *args and **kwargs, so
that it can be general purpose, but you would also like to check the
passed arguments to see if they match a specific function calling
signature.
"""

from inspect import Signature, Parameter

# Create a signature for a func(x, y=42, *, z=None)
params = [
    Parameter("x", Parameter.POSITIONAL_OR_KEYWORD),
    Parameter("y", Parameter.POSITIONAL_OR_KEYWORD, default=42),
    Parameter("z", Parameter.KEYWORD_ONLY, default=None),
]
sig = Signature(params)


def func(*args, **kwargs):
    bound_values = sig.bind(*args, **kwargs)
    for name, value in bound_values.arguments.items():
        print(name, value)
    print()


def make_sig(*names):
    parms = [Parameter(name, Parameter.POSITIONAL_OR_KEYWORD) for name in names]
    return Signature(parms)


class StructureMeta(type):
    def __new__(cls, clsname, bases, clsdict):
        clsdict["__signature__"] = make_sig(*clsdict.get("_fields", []))
        return super().__new__(cls, clsname, bases, clsdict)


class Structure(metaclass=StructureMeta):
    _fields = []

    def __init__(self, *args, **kwargs):
        bound_values = self.__signature__.bind(*args, **kwargs)
        for name, value in bound_values.arguments.items():
            setattr(self, name, value)


# Example
class Stock(Structure):
    _fields = ["name", "shares", "price"]


class Point(Structure):
    _fields = ["x", "y"]


def example_1():
    print("sig:", sig, end="\n\n")
    print("func(1, 2, z=3)")
    func(1, 2, z=3)

    print("func(1)")
    func(1)

    print("func(1, z=3)")
    func(1, z=3)

    print("func(y=2, x=1)")
    func(y=2, x=1)

    print("func(1, 2, 3, 4)")
    try:
        func(1, 2, 3, 4)
    except TypeError as exc:
        print(exc, end="\n\n")

    print("func(y=2)")
    try:
        func(y=2)
    except TypeError as exc:
        print(exc, end="\n\n")


def example_2():
    s1 = Stock("ACME", 100, 490.1)
    print(s1.name, s1.shares, s1.price)

    s2 = Stock(shares=100, name="ACME", price=490.1)
    print(s2.name, s2.shares, s2.price)

    # Not enough args
    try:
        _ = Stock("ACME", 100)
    except TypeError as e:
        print(e)

    # Too many args
    try:
        _ = Stock("ACME", 100, 490.1, "12/21/2012")
    except TypeError as e:
        print(e)

    # Replicated args
    try:
        _ = Stock("ACME", 100, name="ACME", price=490.1)
    except TypeError as e:
        print(e)


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You would like to extract data from a simple XML document.
"""
from xml.etree.ElementTree import parse
from pathlib import Path


def example_1():
    # Download the RSS feed and parse it
    with open(Path(__file__).parent / "sample.xml", "r") as f:
        doc = parse(f)
        print(doc)

    # Extract and output tags of interest
    for item in doc.iterfind("channel/item"):
        title = item.findtext("title")
        date = item.findtext("pubDate")
        link = item.findtext("link")

        print(title)
        print(date)
        print(link)
        print()


def main():
    example_1()


if __name__ == "__main__":
    main()



"""
You want to make a dictionary that maps keys to more than one value
(a so-called 'multidict')
"""
from collections import defaultdict


def main():
    d = {"a": [1, 2, 3], "b": [4, 5]}
    print(d)

    e = {"a": {1, 2, 3}, "b": {4, 5}}
    print(e)

    d = defaultdict(list)
    d["a"].append(1)
    d["a"].append(2)
    d["b"].append(3)
    print(d)

    d = defaultdict(set)
    d["a"].add(1)
    d["a"].add(2)
    d["a"].add(2)
    d["b"].add(3)
    print(d)

    d = {}
    d.setdefault("a", []).append(1)
    d.setdefault("a", []).append(2)
    d.setdefault("a", []).append(3)
    print(d)


if __name__ == "__main__":
    main()



import click
from pyrecipes.cookbook import cookbook
from pyrecipes.errors import ChapterNotFoundError
from pyrecipes.utils.cli import render_chapter


@click.command()
@click.option(
    "-d",
    "--describe",
    is_flag=True,
    help="Shows descriptions of the recipes",
)
@click.argument("chapter", required=False, default=None, type=int)
def ls(chapter, describe):
    """List recipes

    \b
    The default behaviour lists the titles of all recipes for all chapters.
    Limit the output to a specific recipe by adding an additional 'CHAPTER'
    argument e.g. to list chapter 1 recipes

    \b
        recipes ls 1

    You can also add a '-d' flag to display a short description of each recipe.
    """
    if chapter:
        try:
            click.echo(f"{cookbook[chapter].size:,} recipes")
            render_chapter(cookbook[chapter], describe)
        except ChapterNotFoundError as exc:
            click.echo(exc)
    else:
        click.echo(f"{cookbook.size:,} recipes")
        for chapter in cookbook:
            render_chapter(chapter, describe)



"""
You have code that accesses list or tuple elements by position, but this
makes the code difficult to read at time. You'd also like to be less
dependent on position in the structure, by accessing the elements by name.
"""

from collections import namedtuple


def example_1():
    Subscriber = namedtuple("Subscriber", ["email", "joined"])
    print(f"Subscriber named tuple: {Subscriber}")

    sub = Subscriber(email="test@email.com", joined="2023-01-01")
    print(f"sub object: {sub}")
    print(f"email: {sub.email}")
    print(f"joined: {sub.joined}")


def example_2():
    Stock = namedtuple("Stock", ["name", "shares", "price"])

    def compute_cost(records):
        result = 0.0
        for record in records:
            s = Stock(*record)
            print(s)
            result += s.shares * s.price
        return result

    records = [("FB", 10, 12.30), ("MSFT", 20, 101.50), ("AAPL", 15, 120.23)]

    print(f"records: {records}")
    print(compute_cost(records))


def example_3():
    def dict_to_stock(d):
        return stock_prototype._replace(**d)

    Stock = namedtuple("Stock", ["name", "shares", "price", "date", "time"])
    stock_prototype = Stock("", 0, 0.0, None, None)
    print(f"Stock prototype: {stock_prototype}")

    d = {"name": "ACME", "shares": 100, "price": 123.45}
    print(f"dictionary value: {d}")
    print(f"dict to stock: {dict_to_stock(d)}")


def main():
    example_1()
    print("=" * 50)
    example_2()
    print("=" * 50)
    example_3()


if __name__ == "__main__":
    main()



"""
Your code for interacting with the latest web authentication schema
has encountered a singularity and your only solution is to go around
it in the complex plane. Or maybe you just need to perform some
calculations using complex number.
"""
import cmath


def main():
    a = complex(2, 4)
    b = 3 - 5j
    print(a, b)
    print(a.real, a.imag, a.conjugate())
    print(a + b)
    print(cmath.sqrt(-1))
    print(cmath.sin(a))
    print(cmath.cos(a))
    print(cmath.exp(a))


if __name__ == "__main__":
    main()



"""
You want to match text using the same wildcard patterns as are commonly
used when working in Unix shells (e.g. *.py, Dat[0-9]*.csv etc)
"""

from fnmatch import fnmatch, fnmatchcase


def match(term, pattern, case=False):
    result = fnmatchcase(term, pattern) if case else fnmatch(term, pattern)
    result = f'"{term}" matches "{pattern}": {result}'
    if case:
        result += " (case-sensitive)"
    return result


def main():
    print(match("foo.txt", "*.txt"))
    print(match("foo.txt", "?oo.txt"))
    print(match("Dat45.csv", "Dat[0-9]*.csv"))
    print(match("foo.txt", "*.TXT", case=True))


if __name__ == "__main__":
    main()



"""
You have an N-element tuple or sequence that you want to
unpack into a collection of N variables.
"""


def main():
    p = (4, 5)
    print("p:", p)

    x, y = p
    print(x, y)

    data = ["ACME", 50, 9.1, (2012, 12, 21)]
    print("data:", data)
    name, shares, price, date = data
    print("name:", name)
    print("shares:", shares)
    print("price:", price)
    print("date:", date)


if __name__ == "__main__":
    main()



"""
You have a nested sequence that you want to flatten into a single list of values.
"""
from collections.abc import Iterable


def flatten(items, ignore_types=(str, bytes)):
    for item in items:
        if isinstance(item, Iterable) and not isinstance(item, ignore_types):
            yield from flatten(item, ignore_types)
        else:
            yield item


def main():
    items = [1, 2, [3, 4, [5, 6], 7], 8]
    print(f"Items: {items}")
    print(f"flattened: {list(flatten(items))}")

    print("\niterating over items")
    for x in items:
        print(x)

    print("\niterating over flattened items")
    for x in flatten(items):
        print(x)


if __name__ == "__main__":
    main()



"""
You want to create an entirely new attribute type with some
extra functionality in the form of a descriptor class.
"""


class Integer:
    """Descriptor attribute for an integer type-checked attribute"""

    def __init__(self, name):
        self.name = name

    def __get__(self, instance, cls):
        if instance is None:
            return self
        else:
            return instance.__dict__[self.name]

    def __set__(self, instance, value):
        if not isinstance(value, int):
            print("Expected an int")
        else:
            instance.__dict__[self.name] = value

    def __delete__(self, instance):
        del instance.__dict__[self.name]


class Point:
    x = Integer("x")
    y = Integer("y")

    def __init__(self, x, y):
        self.x = x
        self.y = y


def main():
    p = Point(1, 2)
    print(p.x)
    print(p.y)
    p.x = 6
    print(p.x)
    p.y = 8.2
    del p.x


if __name__ == "__main__":
    main()



"""
You are writing classes where you are repeatedly having to define
property methods that perform common tasks, such as type-checking.
You would like to simplify the code so there is not so much code
repetition.
"""


def typed_property(name, expected_type):
    storage_name = "_" + name

    @property
    def prop(self):
        return getattr(self, storage_name)

    @prop.setter
    def prop(self, value):
        if not isinstance(value, expected_type):
            raise TypeError("{} must be a {}".format(name, expected_type))
        setattr(self, storage_name, value)

    return prop


# Example use
class Person:
    name = typed_property("name", str)
    age = typed_property("age", int)

    def __init__(self, name, age):
        self.name = name
        self.age = age


def main():
    p = Person("Dave", 39)
    p.name = "Guido"
    try:
        p.age = "Old"
    except TypeError as e:
        print(e)


if __name__ == "__main__":
    main()



"""
You need to read complicated binary-encoded data that contains a
collection of nested and/or variable-sized records. Such data might
include images, video, shapefiles and so on.
"""
import struct
from pathlib import Path


class StructField:
    def __init__(self, format, offset):
        self.format = format
        self.offset = offset

    def __get__(self, instance, cls):
        if instance is None:
            return self
        else:
            r = struct.unpack_from(self.format, instance._buffer, self.offset)
            return r[0] if len(r) == 1 else r


class StructureMeta(type):
    """
    Metaclass that automatically creates StructField descriptors
    """

    def __init__(self, clsname, bases, clsdict):
        fields = getattr(self, "_fields_", [])
        byte_order = ""
        offset = 0
        for format, fieldname in fields:
            if format.startswith(("<", ">", "!", "@")):
                byte_order = format[0]
                format = format[1:]
            format = byte_order + format
            setattr(self, fieldname, StructField(format, offset))
            offset += struct.calcsize(format)
        setattr(self, "struct_size", offset)


class Structure(metaclass=StructureMeta):
    def __init__(self, bytedata):
        self._buffer = memoryview(bytedata)

    @classmethod
    def from_file(cls, f):
        return cls(f.read(cls.struct_size))


class PolyHeader(Structure):
    file_code = StructField("<i", 0)
    min_x = StructField("<d", 4)
    min_y = StructField("<d", 12)
    max_x = StructField("<d", 20)
    max_y = StructField("<d", 28)
    num_polys = StructField("<i", 36)


def example_1():
    with open(Path(__file__).parent / "polys.bin", "rb") as f:
        data = f.read()

    phead = PolyHeader(data)
    print(phead.file_code == 0x1234)
    print("min_x=", phead.min_x)
    print("max_x=", phead.max_x)
    print("min_y=", phead.min_y)
    print("max_y=", phead.max_y)
    print("num_polys=", phead.num_polys)


def example_2():
    class PolyHeader(Structure):
        _fields_ = [
            ("<i", "file_code"),
            ("d", "min_x"),
            ("d", "min_y"),
            ("d", "max_x"),
            ("d", "max_y"),
            ("i", "num_polys"),
        ]

    with open(Path(__file__).parent / "polys.bin", "rb") as f:
        phead = PolyHeader.from_file(f)
    print(phead.file_code == 0x1234)
    print("min_x=", phead.min_x)
    print("max_x=", phead.max_x)
    print("min_y=", phead.min_y)
    print("max_y=", phead.max_y)
    print("num_polys=", phead.num_polys)


def example_3():
    class StructField:
        """
        Descriptor representing a simple structure field
        """

        def __init__(self, format, offset):
            self.format = format
            self.offset = offset

        def __get__(self, instance, cls):
            if instance is None:
                return self
            else:
                r = struct.unpack_from(self.format, instance._buffer, self.offset)
                return r[0] if len(r) == 1 else r

    class NestedStruct:
        """
        Descriptor representing a nested structure
        """

        def __init__(self, name, struct_type, offset):
            self.name = name
            self.struct_type = struct_type
            self.offset = offset

        def __get__(self, instance, cls):
            if instance is None:
                return self
            else:
                data = instance._buffer[
                    self.offset : self.offset + self.struct_type.struct_size
                ]
                result = self.struct_type(data)
                setattr(instance, self.name, result)
                return result

    class StructureMeta(type):
        """
        Metaclass that automatically creates StructField descriptors
        """

        def __init__(self, clsname, bases, clsdict):
            fields = getattr(self, "_fields_", [])
            byte_order = ""
            offset = 0
            for format, fieldname in fields:
                if isinstance(format, StructureMeta):
                    setattr(self, fieldname, NestedStruct(fieldname, format, offset))
                    offset += format.struct_size
                else:
                    if format.startswith(("<", ">", "!", "@")):
                        byte_order = format[0]
                        format = format[1:]
                    format = byte_order + format
                    setattr(self, fieldname, StructField(format, offset))
                    offset += struct.calcsize(format)
            setattr(self, "struct_size", offset)

    class Structure(metaclass=StructureMeta):
        def __init__(self, bytedata):
            self._buffer = memoryview(bytedata)

        @classmethod
        def from_file(cls, f):
            return cls(f.read(cls.struct_size))

    class Point(Structure):
        _fields_ = [("<d", "x"), ("d", "y")]

    class PolyHeader(Structure):
        _fields_ = [
            ("<i", "file_code"),
            (Point, "min"),
            (Point, "max"),
            ("i", "num_polys"),
        ]

    with open(Path(__file__).parent / "polys.bin", "rb") as f:
        phead = PolyHeader.from_file(f)
    print(phead.file_code == 0x1234)
    print("min.x=", phead.min.x)
    print("max.x=", phead.max.x)
    print("min.y=", phead.min.y)
    print("max.y=", phead.max.y)
    print("num_polys=", phead.num_polys)


def main():
    example_1()
    example_2()
    example_3()


if __name__ == "__main__":
    main()



"""
You want to add extra processing (e.g. type checking or validation) to
the getting or setting of an instance attribute.
"""

import math


class Person:
    def __init__(self, first_name):
        self.first_name = first_name

    @property
    def first_name(self):
        return self._first_name

    @first_name.setter
    def first_name(self, value):
        if not isinstance(value, str):
            raise TypeError("Expected a string.")
        self._first_name = value

    @first_name.deleter
    def first_name(self):
        raise AttributeError("Can't delete attribute.")


class Circle:
    def __init__(self, radius):
        self.radius = radius

    @property
    def area(self):
        return math.pi * self.radius**2

    @property
    def circumference(self):
        return math.pi * 2 * self.radius


def example_1():
    person = Person("Chris")
    print(person.first_name)

    try:
        person.first_name = 2
    except TypeError as exc:
        print(exc)

    person.first_name = "Bob"
    print(person.first_name)

    try:
        del person.first_name
    except AttributeError as exc:
        print(exc)


def example_2():
    c = Circle(4.0)
    print(c.radius)
    print(c.area)
    print(c.circumference)


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You want to sort objects of the same class, but they don't natively
support comparison operations.
"""

from operator import attrgetter


class User:
    def __init__(self, user_id):
        self.user_id = user_id

    def __repr__(self):
        return f"User({self.user_id})"


def main():
    users = [User(12), User(23), User(2), User(99)]
    print(f"Users: {users}\n")

    print("Sorted using lambda:")
    print(sorted(users, key=lambda u: u.user_id), end="\n\n")

    print("Sorted using attrgetter:")
    print(sorted(users, key=attrgetter("user_id")))


if __name__ == "__main__":
    main()



"""
You want to add of change the Unicode encoding of an already open file
without closing it first.
"""
import io
import sys
import urllib.request


def example_1():
    """For illustrative purposes only"""
    u = urllib.request.urlopen("http://www.python.org")
    f = io.TextIOWrapper(u, encoding="utf-8")
    text = f.read()
    print(text)


def example_2():
    """For illustrative purposes only"""
    print(sys.stdout.encoding)
    sys.stdout = io.TextIOWrapper(sys.stdout.detatch(), encoding="latin-1")
    print(sys.stdout.encoding)


def main():
    pass


if __name__ == "__main__":
    main()



"""
You want to read or write data encoded as a CSV file.
"""

import csv
from pathlib import Path
from collections import namedtuple


def example_1():
    with open(Path(__file__).parent / "stocks.csv", "r") as f:
        f_csv = csv.reader(f)
        print(f_csv)
        headers = next(f_csv)
        print(f"Header: {headers}")
        for i, row in enumerate(f_csv, 1):
            print(f"[{i:^3}] - {row}")


def example_2():
    with open(Path(__file__).parent / "stocks.csv", "r") as f:
        f_csv = csv.reader(f)
        headers = next(f_csv)
        Row = namedtuple("Row", headers)
        print(headers)
        for r in f_csv:
            row = Row(*r)
            print(row)


def example_3():
    with open(Path(__file__).parent / "stocks.csv", "r") as f:
        f_csv = csv.DictReader(f)
        for r in f_csv:
            print(r)


def main():
    example_1()
    example_2()
    example_3()


if __name__ == "__main__":
    main()



"""
Your program creates data structures with cycles (e.g. trees, graphs,
observer patterns, etc.), but you are experiencing problems with
memory management.
"""
import weakref


class Node:
    def __init__(self, value):
        self.value = value
        self._parent = None
        self.children = []

    def __repr__(self):
        return "Node({!r:})".format(self.value)

    # property that manages the parent as a weak-reference
    @property
    def parent(self):
        return self._parent if self._parent is None else self._parent()

    @parent.setter
    def parent(self, node):
        self._parent = weakref.ref(node)

    def add_child(self, child):
        self.children.append(child)
        child.parent = self


def main():
    root = Node("parent")
    c1 = Node("c1")
    c2 = Node("c2")
    root.add_child(c1)
    root.add_child(c2)

    print(c1.parent)
    del root
    print(c1.parent)


if __name__ == "__main__":
    main()



"""
You want to organise your code into a package consisting of a hierarchical
collection of modules.
"""

from pathlib import Path
import textwrap
from pyrecipes.utils.text import tree


def main():
    print(
        textwrap.dedent(
            """
            To define a python package, simply organise your code on your file system
            into a sensible directory/file structure and add an '__init__.py' file
            in every directory.

            For example, here would be a structure for a 'graphics' package,
            with subpackages 'primitive' with modules 'line', 'fill' & 'text'
            and subpackage 'formats' with modules 'png' & 'jpg':

            """
        )
    )

    for line in tree(Path(__file__).parent / "src"):
        print(line)
    print(
        textwrap.dedent(
            """

            With this done you should be able to perform various import statements
            such as the following:

            import graphics.primitive.line
            from graphics.primitive import line
            import graphics.formats.jpg as jpg
            """
        )
    )


if __name__ == "__main__":
    main()



"""
You have entered a time machine and suddenly find yourself working
on elemntary-level homework problems involving fractions. Or perhaps
you're writing code to make calculations involving measurements
made in your wood shop.
"""
from fractions import Fraction


def main():
    a = Fraction(5, 4)
    b = Fraction(7, 16)
    print("a:", a, "b:", b)
    print("a + b:", a + b)
    print("a - b:", a - b)
    print("a * b:", a * b)
    print("a / b:", a / b)

    c = a * b
    print(c)
    print(c.numerator)
    print(c.denominator)
    print(float(c))
    print(c.limit_denominator(8))

    x = 3.75
    y = Fraction(*x.as_integer_ratio())
    print(x, y)


if __name__ == "__main__":
    main()



"""
You need to perform the same operation on many objects, bu the objects are
contained in different containers, and you'd like to avoid nested loops
without losing the readability of your code.
"""
from itertools import chain


def main():
    a = [1, 2, 3, 4]
    b = ["x", "y", "z"]

    print("a:", a, " | b:", b)
    for x in chain(a, b):
        print(x)

    # Various working sets of items
    active_items = set([1, 2, 3])
    inactive_items = set([4, 5, 6])

    for item in chain(active_items, inactive_items):
        # process item
        print(f"processing... {item}")


if __name__ == "__main__":
    main()



"""
You want to define a function or method where one or more of the
arguments are optional and have a default value.
"""


def spam(a, b=42):
    print(a, b)


def spam2(a, b=None):
    b = [] if b is None else b
    print(a, b)


def main():
    spam(1)
    spam(1, 2)
    spam2(1)
    spam2(2, [1, 2, 3])


if __name__ == "__main__":
    main()



"""
Within a subclass, you want to extend the functionality of a property
defined in a parent class.
"""


class Person:
    def __init__(self, name):
        self.name = name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            raise TypeError("Expected a string.")
        self._name = value

    @name.deleter
    def name(self):
        print("Cannot delete attribute.")


class SubPerson(Person):
    @property
    def name(self):
        print("getting name")
        return super().name

    @name.setter
    def name(self, value):
        print(f"setting name to {value}")
        super(SubPerson, SubPerson).name.__set__(self, value)

    @name.deleter
    def name(self):
        print("Deleting name.")
        super(SubPerson, SubPerson).name.__delete__(self)


def main():
    sub_person = SubPerson("Chris")
    print(sub_person.name)
    sub_person.name = "Bob"
    del sub_person.name


if __name__ == "__main__":
    main()



"""
You want to write raw bytes to a file opened in text mode.
"""
import sys


def main():
    try:
        sys.stdout.write(b"Hello\n")
    except TypeError as exc:
        print("Caught exception:", exc)

    sys.stdout.buffer.write(b"Hello\n")


if __name__ == "__main__":
    main()



"""
You have some code that needs to loop over each date in the current
month and want an efficient way to calculate that date range.
"""
from datetime import datetime, date, timedelta
import calendar


def get_month_range(start_date=None):
    if start_date is None:
        start_date = date.today().replace(day=1)
    _, days_in_month = calendar.monthrange(start_date.year, start_date.month)
    end_date = start_date + timedelta(days=days_in_month)
    return start_date, end_date


def date_range(start, stop, step):
    while start < stop:
        yield start
        start += step


def main():
    a_day = timedelta(days=1)
    first_day, last_day = get_month_range()

    while first_day < last_day:
        print(first_day)
        first_day += a_day

    print()
    for d in date_range(datetime(2023, 1, 1), datetime(2023, 1, 5), timedelta(hours=6)):
        print(d)


if __name__ == "__main__":
    main()



"""
You want to write programs that parse and analyze Python source code.
"""

import ast
import time
import inspect


class CodeAnalyzer(ast.NodeVisitor):
    def __init__(self):
        self.loaded = set()
        self.stored = set()
        self.deleted = set()

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Load):
            self.loaded.add(node.id)
        elif isinstance(node.ctx, ast.Store):
            self.stored.add(node.id)
        elif isinstance(node.ctx, ast.Del):
            self.deleted.add(node.id)


# Some python code
code = """
for i in range(10):
    print(i)
del i
"""


# Node visitor that lowers globally accessed names into
# the function body as local variables.
class NameLower(ast.NodeVisitor):
    def __init__(self, lowered_names):
        self.lowered_names = lowered_names

    def visit_FunctionDef(self, node):
        # Compile some assignments to lower the constants
        code = "__globals = globals()\n"
        code += "\n".join(
            "{0} = __globals['{0}']".format(name) for name in self.lowered_names
        )

        code_ast = ast.parse(code, mode="exec")

        # Inject new statements into the function body
        node.body[:0] = code_ast.body

        # Save the function object
        self.func = node


# Decorator that turns global names into locals
def lower_names(*namelist):
    def lower(func):
        srclines = inspect.getsource(func).splitlines()
        # Skip source lines prior to the @lower_names decorator
        for n, line in enumerate(srclines):
            if "@lower_names" in line:
                break

        src = "\n".join(srclines[n + 1 :])
        # Hack to deal with indented code
        if src.startswith((" ", "\t")):
            src = "if 1:\n" + src
        top = ast.parse(src, mode="exec")

        # Transform the AST
        cl = NameLower(namelist)
        cl.visit(top)

        # Execute the modified AST
        temp = {}
        exec(compile(top, "", "exec"), temp, temp)

        # Pull out the modified code object
        func.__code__ = temp[func.__name__].__code__
        return func

    return lower


# Example of use
INCR = 1


def countdown1(n):
    while n > 0:
        n -= INCR


@lower_names("INCR")
def countdown2(n):
    while n > 0:
        n -= INCR


def example_1():
    # Parse into an AST
    top = ast.parse(code, mode="exec")

    # Feed the AST to analyze name usage
    c = CodeAnalyzer()
    c.visit(top)
    print("Loaded:", c.loaded)
    print("Stored:", c.stored)
    print("Deleted:", c.deleted)
    print()


def example_2():
    print("Running a performance check")

    start = time.time()
    countdown1(100)
    end = time.time()
    print("countdown1:", end - start)

    start = time.time()
    countdown2(100)
    end = time.time()
    print("countdown2:", end - start)


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You need to parse text according to a set of grammar rules and perform
actions or build an abstract syntax tree representing the input. The
grammar is small, so you'd prefer to just write the parser yourself
as opposed to using some kind of framework.
"""
import re
import collections

# Token specification
NUM = r"(?P<NUM>\d+)"
PLUS = r"(?P<PLUS>\+)"
MINUS = r"(?P<MINUS>-)"
TIMES = r"(?P<TIMES>\*)"
DIVIDE = r"(?P<DIVIDE>/)"
LPAREN = r"(?P<LPAREN>\()"
RPAREN = r"(?P<RPAREN>\))"
WS = r"(?P<WS>\s+)"

master_pat = re.compile("|".join([NUM, PLUS, MINUS, TIMES, DIVIDE, LPAREN, RPAREN, WS]))

# Tokenizer
Token = collections.namedtuple("Token", ["type", "value"])


def generate_tokens(text):
    scanner = master_pat.scanner(text)
    for m in iter(scanner.match, None):
        tok = Token(m.lastgroup, m.group())
        if tok.type != "WS":
            yield tok


# Parser
class ExpressionEvaluator:
    """
    Implementation of a recursive descent parser.   Each method
    implements a single grammar rule.  Use the ._accept() method
    to test and accept the current lookahead token.  Use the ._expect()
    method to exactly match and discard the next token on on the input
    (or raise a SyntaxError if it doesn't match).
    """

    def parse(self, text):
        self.tokens = generate_tokens(text)
        self.tok = None  # Last symbol consumed
        self.nexttok = None  # Next symbol tokenized
        self._advance()  # Load first lookahead token
        return self.expr()

    def _advance(self):
        "Advance one token ahead"
        self.tok, self.nexttok = self.nexttok, next(self.tokens, None)

    def _accept(self, toktype):
        "Test and consume the next token if it matches toktype"
        if self.nexttok and self.nexttok.type == toktype:
            self._advance()
            return True
        else:
            return False

    def _expect(self, toktype):
        "Consume next token if it matches toktype or raise SyntaxError"
        if not self._accept(toktype):
            raise SyntaxError("Expected " + toktype)

    # Grammar rules follow

    def expr(self):
        "expression ::= term { ('+'|'-') term }*"

        exprval = self.term()
        while self._accept("PLUS") or self._accept("MINUS"):
            op = self.tok.type
            right = self.term()
            if op == "PLUS":
                exprval += right
            elif op == "MINUS":
                exprval -= right
        return exprval

    def term(self):
        "term ::= factor { ('*'|'/') factor }*"

        termval = self.factor()
        while self._accept("TIMES") or self._accept("DIVIDE"):
            op = self.tok.type
            right = self.factor()
            if op == "TIMES":
                termval *= right
            elif op == "DIVIDE":
                termval /= right
        return termval

    def factor(self):
        "factor ::= NUM | ( expr )"

        if self._accept("NUM"):
            return int(self.tok.value)
        elif self._accept("LPAREN"):
            exprval = self.expr()
            self._expect("RPAREN")
            return exprval
        else:
            raise SyntaxError("Expected NUMBER or LPAREN")


def example_1():
    e = ExpressionEvaluator()
    print(e.parse("2"))
    print(e.parse("2 + 3"))
    print(e.parse("2 + 3 * 4"))
    print(e.parse("2 + (3 + 4) * 5"))


class ExpressionTreeBuilder(ExpressionEvaluator):
    def expr(self):
        "expression ::= term { ('+'|'-') term }"

        exprval = self.term()
        while self._accept("PLUS") or self._accept("MINUS"):
            op = self.tok.type
            right = self.term()
            if op == "PLUS":
                exprval = ("+", exprval, right)
            elif op == "MINUS":
                exprval = ("-", exprval, right)
        return exprval

    def term(self):
        "term ::= factor { ('*'|'/') factor }"

        termval = self.factor()
        while self._accept("TIMES") or self._accept("DIVIDE"):
            op = self.tok.type
            right = self.factor()
            if op == "TIMES":
                termval = ("*", termval, right)
            elif op == "DIVIDE":
                termval = ("/", termval, right)
        return termval

    def factor(self):
        "factor ::= NUM | ( expr )"

        if self._accept("NUM"):
            return int(self.tok.value)
        elif self._accept("LPAREN"):
            exprval = self.expr()
            self._expect("RPAREN")
            return exprval
        else:
            raise SyntaxError("Expected NUMBER or LPAREN")


# Example of building trees
def example_2():
    e = ExpressionTreeBuilder()
    print(e.parse("2 + 3"))
    print(e.parse("2 + 3 * 4"))
    print(e.parse("2 + (3 + 4) * 5"))
    print(e.parse("2 + 3 + 4"))


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You want to implement a priority queue that sorts items by a given priority and always returns
the item with the highest priority on each pop operation.
"""

import heapq


class PriorityQueue:
    def __init__(self):
        self._queue = []
        self._index = 0

    def push(self, item, priority):
        heapq.heappush(self._queue, (-priority, self._index, item))
        self._index += 1

    def pop(self):
        return heapq.heappop(self._queue)[-1]

    def __repr__(self):
        return f"PriorityQueue({len(self._queue)} items)"

    @property
    def size(self):
        return len(self._queue)


class Item:
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f"Item({self.name})"


def main():
    print("instantiating queue")
    q = PriorityQueue()
    print(q)
    items = [
        (Item("fourth"), 1),
        (Item("first"), 5),
        (Item("third"), 2),
        (Item("second"), 4),
    ]
    for item, priority in items:
        print(f"Adding item {item} with priority {priority}")
        q.push(item, priority)
    print(q)
    print("\npopping items in order of priority:")
    for i in range(q.size):
        item = q.pop()
        print(f"Popped {item}")


if __name__ == "__main__":
    main()



"""
You have a string of text that you want to parse left to right
into a stream of tokens.
"""
import re
from collections import namedtuple

NAME = r"(?P<NAME>[a-zA-Z_][a-zA-Z_0-9]*)"
NUM = r"(?P<NUM>\d+)"
PLUS = r"(?P<PLUS>\+)"
TIMES = r"(?P<TIMES>\*)"
EQ = r"(?P<EQ>=)"
WS = r"(?P<WS>\s+)"

Token = namedtuple("Token", ["type", "value"])


def generate_tokens(pat, text):
    scanner = pat.scanner(text)
    for m in iter(scanner.match, None):
        yield Token(m.lastgroup, m.group())


def main():
    text = "foo = 23 + 42 * 10"
    master_pat = re.compile("|".join([NAME, NUM, PLUS, TIMES, EQ, WS]))
    scanner = master_pat.scanner(text)

    print(text)
    print(master_pat)
    for i in range(3):
        m = scanner.match()
        print(m, m.lastgroup, m.group())

    for tok in generate_tokens(master_pat, text):
        print(tok)


if __name__ == "__main__":
    main()



"""
You are writing a lot of classes that serve as data structures, but are
getting tired of writing highly repetitive and boilerplate __init__()
functions.
"""

import math


class Structure:
    # class variable that specifies expected fields
    _fields = []

    def __init__(self, *args, **kwargs):
        n = len(self._fields)
        if len(args) != n:
            raise TypeError(f"Expected {n} fields")

        # Set required arguments
        for name, value in zip(self._fields, args):
            setattr(self, name, value)

        # Set additional arguments if any
        extra_args = kwargs.keys() - self._fields
        for name in extra_args:
            setattr(self, name, kwargs.pop(name))
        if kwargs:
            raise TypeError(f'Duplicated values for {",".join(kwargs)}')


class Stock(Structure):
    _fields = ["name", "shares", "price"]


class Point(Structure):
    _fields = ["x", "y"]


class Circle(Structure):
    _fields = ["radius"]

    @property
    def area(self):
        return math.pi * self.radius**2


def main():
    s1 = Stock("AAPL", 10, 100)
    s2 = Stock("MSFT", 10, 95, date="2023-01-01")
    c = Circle(4.0)
    print(s1.name)
    print(s1.price)
    print(s2.date)
    print(c.radius)
    print(c.area)


if __name__ == "__main__":
    main()



"""
You need to supply a short callback function for use with an operation
such as sort(), but you don't want to write separate one-line function
using the def statement. Instead you'd like a shortcut that allows you
to specify the function "in line".
"""


def main():
    add = lambda x, y: x + y
    print(add(2, 3))
    print(add("Hello ", "World!"))

    names = ["David Beazley", "Brian Jones", "Raymond Hattinger", "Ned Batchelder"]
    print("names:", names)
    print("sorted names:", sorted(names, key=lambda x: x.split()[-1].lower()))


if __name__ == "__main__":
    main()



"""
You want to iterate over all the possible combinations or permutations of a collection of items.
"""
from itertools import permutations, combinations


def main():
    items = ["a", "b", "c"]
    print(f"all permutations of {items}")
    for p in permutations(items):
        print(p)

    print(f"\nall permutations of {items} of length 2")
    for p in permutations(items, 2):
        print(p)

    print(f"all combinations of {items} of length 3")
    for c in combinations(items, 3):
        print(c)

    print(f"\nall combinations of {items} of length 2")
    for c in combinations(items, 2):
        print(c)

    print(f"\nall combinations of {items} of length 1")
    for c in combinations(items, 1):
        print(c)


if __name__ == "__main__":
    main()



"""
You want to initialize parts of a class definition once at the time
a class is defined, not when instances are created.
"""

import operator


class StructTupleMeta(type):
    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for n, name in enumerate(cls._fields_):
            setattr(cls, name, property(operator.itemgetter(n)))


class StructTuple(tuple, metaclass=StructTupleMeta):
    _fields_ = []

    def __new__(cls, *args):
        if len(args) != len(cls._fields_):
            raise ValueError("{} arguments required".format(len(cls._fields_)))
        return super().__new__(cls, args)


# Examples
class Stock(StructTuple):
    _fields_ = ["name", "shares", "price"]


class Point(StructTuple):
    _fields_ = ["x", "y"]


def main():
    s = Stock("ACME", 50, 91.1)
    print(s)
    print(s[0])
    print(s.name)
    print(s.shares * s.price)
    try:
        s.shares = 23
    except AttributeError as e:
        print(e)


if __name__ == "__main__":
    main()



"""
You need to decode a string of hexadecimal digits into a byte string
or encode a byte string as hex.
"""
import binascii
import base64


def main():
    s = b"hello"
    h = binascii.b2a_hex(s)
    print("s:", s)
    print("h:", h)

    print(binascii.a2b_hex(h))

    h = base64.b16encode(s)
    print(h)
    print(base64.b16decode(h))


if __name__ == "__main__":
    main()



"""
You want to memory map a binary file into a mutable byte array, possibly
for random access to it's contents of in-place modifications.
"""
import os
import mmap
from pathlib import Path


def memory_map(filename, access=mmap.ACCESS_WRITE):
    size = os.path.getsize(filename)
    fd = os.open(filename, os.O_RDWR)
    return mmap.mmap(fd, size, access=access)


def main():
    size = 100000
    filename = Path(__file__).parent / "somefile.data"
    with filename.open("wb") as f:
        f.seek(size - 1)
        f.write(b"\x00")

    with memory_map(filename) as m:
        print(m)
        print(len(m))
        print(m[:10])
        m[0:11] = b"Hello World"
        print("memorymap closed:", m.closed)
    print("memorymap closed:", m.closed)

    with filename.open("rb") as f:
        print(f.read(15))


if __name__ == "__main__":
    main()



"""
You want to automatically record the order in which attributes
and methods are defined inside a class body so that you can use it
in various operations (e.g. serializing, mapping to databases etc.)
"""

from collections import OrderedDict


class Typed:
    _expected_type = type(None)

    def __init__(self, name=None):
        self._name = name

    def __set__(self, instance, value):
        if not isinstance(value, self._expected_type):
            raise TypeError(f"Expected {self._expected_type} - got {type(value)}")
        instance.__dict__[self._name] = value


class Integer(Typed):
    _expected_type = int


class Float(Typed):
    _expected_type = float


class String(Typed):
    _expected_type = str


# Metaclass that uses an OrderedDict for class body
class OrderedMeta(type):
    def __new__(cls, clsname, bases, clsdict):
        d = dict(clsdict)
        order = []
        for name, value in clsdict.items():
            if isinstance(value, Typed):
                value._name = name
                order.append(name)
        d["_order"] = order
        return type.__new__(cls, clsname, bases, d)

    @classmethod
    def __prepare__(cls, clsname, bases):
        return OrderedDict()


class Structure(metaclass=OrderedMeta):
    def as_csv(self):
        return ",".join(str(getattr(self, name)) for name in self._order)


class Stock(Structure):
    name = String()
    shares = Integer()
    price = Float()

    def __init__(self, name, shares, price):
        self.name = name
        self.shares = shares
        self.price = price


def main():
    stock = Stock("GOOG", 100, 490.1)
    print(stock)
    print(stock.name)
    print(stock.shares)
    print(stock.price)
    print(stock.as_csv())
    print(stock._order)

    try:
        stock = Stock("GOOG", "a lot", 490.1)
    except TypeError as exc:
        print('Tried to instantiate: Stock("GOOG", "a lot", 490.1)')
        print(exc)


if __name__ == "__main__":
    main()



"""
You want to implement a custom class that mimics the behaviour of a
common built-in container type such as a list or dictionary. However,
you are not entirely sure what methods need to be implemented to do it.
"""
import collections
import bisect


class A(collections.abc.Iterable):
    # Required dunder-method to inherit from Iterable
    def __iter__(self):
        pass

    def __repr__(self):
        return "A()"


class SortedItems(collections.abc.Sequence):
    def __init__(self, initial=None):
        self._items = sorted(initial) if initial is not None else []

    # Required for Sequence
    def __getitem__(self, index):
        return self._items[index]

    # Required for Sequence
    def __len__(self):
        return len(self._items)

    def add(self, item):
        bisect.insort(self._items, item)

    def __repr__(self):
        return f"{self.__class__.__name__}({self._items})"


class Items(collections.abc.MutableSequence):
    pass


def main():
    a = A()
    print(a)

    sorted_items = SortedItems([3, 2, 5])
    print(sorted_items)
    print(len(sorted_items))
    sorted_items.add(1)
    print(sorted_items)
    print(len(sorted_items))

    try:
        Items()
    except TypeError as exc:
        print(exc)


if __name__ == "__main__":
    main()



"""
You need to process items in an iterator, but for whatever reason
you can't or don't want to use a for loop.
"""

from pathlib import Path


def main():
    with open(Path(__file__).parent / "somefile.txt", "r") as f:
        try:
            while True:
                line = next(f)
                print(line.rstrip("\n"), end="")
        except StopIteration:
            pass


if __name__ == "__main__":
    main()



"""
You need to split a string into fields, but the delimiters (and spacing around them) aren't consitent
thoughout the string.
"""

import re

line = "asdf fjdk; afed, fjek,asdf       foo"


def example_1():
    pat = r"[;,\s]\s*"
    print("Example 1")
    print(f"Splitting the line:\n\t{line}")
    print(f'using the regex pattern "{pat}"\n')
    print(f"Fields: {re.split(pat, line)}")
    print("=" * 50)


def example_2():
    pat = r"([;,\s])\s*"
    print("Example 2")
    print(f"Splitting the line:\n\t{line}")
    print(f'using the regex pattern "{pat}"')
    print("notice this time how the regex pattern contains a capture group\n")
    fields = re.split(pat, line)
    values = fields[::2]
    delimiters = fields[1::2] + [""]
    print(f"Fields: {fields}")
    print(f"Values: {values}")
    print(f"Delimiters: {delimiters}")
    print(
        f'Reconstructed example: {"".join(v + d for v, d in zip(values, delimiters))}'
    )
    print("=" * 50)


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You want to define various kinds of data structures, but want to enforce
contraints on the values that are allowed to be assigned to certain attributes.
"""


# Base class. Uses a descriptor to set a value
class Descriptor:
    def __init__(self, name=None, **opts):
        self.name = name
        self.__dict__.update(opts)

    def __set__(self, instance, value):
        instance.__dict__[self.name] = value


# Descriptor for enforcing types
class Typed(Descriptor):
    expected_type = type(None)

    def __set__(self, instance, value):
        if not isinstance(value, self.expected_type):
            raise TypeError("expected " + str(self.expected_type))
        super().__set__(instance, value)


# Descriptor for enforcing values
class Unsigned(Descriptor):
    def __set__(self, instance, value):
        if value < 0:
            raise ValueError("Expected >= 0")
        super().__set__(instance, value)


class MaxSized(Descriptor):
    def __init__(self, name=None, **opts):
        if "size" not in opts:
            raise TypeError("missing size option")
        self.size = opts["size"]
        super().__init__(name, **opts)

    def __set__(self, instance, value):
        if len(value) >= self.size:
            raise ValueError("size must be < " + str(self.size))
        super().__set__(instance, value)


class Integer(Typed):
    expected_type = int


class UnsignedInteger(Integer, Unsigned):
    pass


class Float(Typed):
    expected_type = float


class UnsignedFloat(Float, Unsigned):
    pass


class String(Typed):
    expected_type = str


class SizedString(String, MaxSized):
    pass


# Class decorator to apply constraints
def check_attributes(**kwargs):
    def decorate(cls):
        for key, value in kwargs.items():
            if isinstance(value, Descriptor):
                value.name = key
                setattr(cls, key, value)
            else:
                setattr(cls, key, value(key))
        return cls

    return decorate


# A metaclass that applies checking
class checkedmeta(type):
    def __new__(cls, clsname, bases, methods):
        # Attach attribute names to the descriptors
        for key, value in methods.items():
            if isinstance(value, Descriptor):
                value.name = key
        return type.__new__(cls, clsname, bases, methods)


# Testing code
def test(s):
    print(s.name)
    s.shares = 75
    print(s.shares)
    try:
        s.shares = -10
    except ValueError as e:
        print(e)
    try:
        s.price = "a lot"
    except TypeError as e:
        print(e)

    try:
        s.name = "ABRACADABRA"
    except ValueError as e:
        print(e)


def main():
    print("# --- Class with descriptors")

    class Stock:
        # Specify constraints
        name = SizedString("name", size=8)
        shares = UnsignedInteger("shares")
        price = UnsignedFloat("price")

        def __init__(self, name, shares, price):
            self.name = name
            self.shares = shares
            self.price = price

    s = Stock("ACME", 50, 91.1)
    test(s)

    print("# --- Class with class decorator")

    @check_attributes(
        name=SizedString(size=8), shares=UnsignedInteger, price=UnsignedFloat
    )
    class Stock:
        def __init__(self, name, shares, price):
            self.name = name
            self.shares = shares
            self.price = price

    s = Stock("ACME", 50, 91.1)
    test(s)

    print("# --- Class with metaclass")

    class Stock(metaclass=checkedmeta):
        name = SizedString(size=8)
        shares = UnsignedInteger()
        price = UnsignedFloat()

        def __init__(self, name, shares, price):
            self.name = name
            self.shares = shares
            self.price = price

    s = Stock("ACME", 50, 91.1)
    test(s)


if __name__ == "__main__":
    main()



"""
You've written a decorator, but when you appl it to a function, important
metadata such as the name, doc string, annotations and calling signature
are lost.
"""

import time
from functools import wraps


def timethis(func):
    """Decorator that reports the execution time."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(func.__name__, end - start)
        return result

    return wrapper


@timethis
def countdown(n: int):
    """Counts down"""
    while n > 0:
        n -= 1


def main():
    countdown(100000)
    print("Name:", countdown.__name__)
    print("Docstring:", repr(countdown.__doc__))
    print("Annotations:", countdown.__annotations__)


if __name__ == "__main__":
    main()



"""
You have multiple threads in your program and you want
to safely communicate or exchange data between them.
"""

import random
import threading
import heapq
from queue import Queue
from threading import Thread

_sentinel = object()


def producer(out_q):
    """A thread that produces data."""
    random.seed(2023)
    while True:
        # Produce some data ...
        data = random.randint(1, 100)
        out_q.put(data)
        print(f"produced: {data}")
        if data >= 95:
            print("Stopping.")
            out_q.put(_sentinel)
            break


def consumer(in_q):
    """A thread that consumes data"""
    while True:
        # Get some data
        data = in_q.get()

        if data is _sentinel:
            in_q.put(_sentinel)
            print("Stopping")
            break

        # Do some processing...
        print(f"consumed: {data}")


class ThreadSafePriorityQueue:
    def __init__(self) -> None:
        self._queue = []
        self._count = 0
        self._cv = threading.Condition()

    def put(self, item, priority):
        with self._cv:
            heapq.heappush(self._queue, (-priority, self._count, item))
            self._count += 1
            self._cv.notify()

    def get(self):
        with self._cv:
            while len(self._queue) == 0:
                self._cv.wait()
            return heapq.heappop(self._queue)[-1]


def example_1():
    q = Queue()
    t1 = Thread(target=consumer, args=(q,))
    t2 = Thread(target=producer, args=(q,))
    t1.start()
    t2.start()


def main():
    example_1()


if __name__ == "__main__":
    main()



"""
You want to put a wrapper layer around a function that adds extra processing (e.g. logging, timing etc.)
"""
import time
from functools import wraps


def timethis(func):
    """
    Decorator that reports the execution time.
    """

    def wrapper(*args, **kwargs):
        """wrapped func"""
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(func.__name__, end - start)
        return result

    return wrapper


def timethis_v2(func):
    """
    Decorator that reports the execution time.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        """wrapped func"""
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(func.__name__, end - start)
        return result

    return wrapper


@timethis
def countdown(n: int):
    """Counts down from n"""
    while n > 0:
        n -= 1


@timethis_v2
def countdown_v2(n: int):
    """Counts down from n"""
    while n > 0:
        n -= 1


def run(func, list_n):
    for n in list_n:
        func(n)
    print(f"function name: {func.__name__}")
    print(f"function docstring: {func.__doc__}")
    print(f"function annotations: {func.__annotations__}")


def main():
    print("Without preserving metadata...")
    run(countdown, [10, 1000, 1000000, 10000000])

    print("\nPreserving metadata... (with functools.wraps)")
    run(countdown_v2, [10, 1000, 1000000, 10000000])


if __name__ == "__main__":
    main()



"""
Your program consists of a large class hierarchy and you would like to
enforce certain kinds of coding conventions (or perform diagnostics) to
help maintain programmer sanity.
"""


def example_1():
    class NoMixedCaseMeta(type):
        def __new__(cls, clsname, bases, clsdict):
            for name in clsdict:
                if name.lower() != name:
                    raise TypeError("Bad attribute name: " + name)
            return super().__new__(cls, clsname, bases, clsdict)

    class Root(metaclass=NoMixedCaseMeta):
        pass

    class A(Root):
        def foo_bar(self):  # Ok
            pass

    print("**** About to generate a TypeError")
    try:

        class B(Root):
            def fooBar(self):  # TypeError
                pass

        B()

    except TypeError as exc:
        print(exc)


def example_2():
    from inspect import signature
    import logging

    class MatchSignaturesMeta(type):
        def __init__(self, clsname, bases, clsdict):
            super().__init__(clsname, bases, clsdict)
            sup = super(self, self)
            for name, value in clsdict.items():
                if name.startswith("_") or not callable(value):
                    continue
                # Get the previous definition (if any) and compare the signatures
                prev_dfn = getattr(sup, name, None)
                if prev_dfn:
                    prev_sig = signature(prev_dfn)
                    val_sig = signature(value)
                    if prev_sig != val_sig:
                        logging.warning(
                            "Signature mismatch in %s. %s != %s",
                            value.__qualname__,
                            str(prev_sig),
                            str(val_sig),
                        )

    # Example
    class Root(metaclass=MatchSignaturesMeta):
        pass

    class A(Root):
        def foo(self, x, y):
            pass

        def spam(self, x, *, z):
            pass

    # Class with redefined methods, but slightly different signatures
    class B(A):
        def foo(self, a, b):
            pass

        def spam(self, x, z):
            pass


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You need to execute a reduction function (e.g. sum(), min(), max()), but first need
to transform or filter the data.
"""


def main():
    # Output a tuple as CSV
    s = ("ACME", 50, 123.45)
    print(",".join(str(x) for x in s))

    # Data reduction across fields of a data structure
    portfolio = [
        {"name": "GOOG", "shares": 50},
        {"name": "YHOO", "shares": 75},
        {"name": "AOL", "shares": 20},
        {"name": "SCOX", "shares": 65},
    ]
    min_shares = min(s["shares"] for s in portfolio)
    print(min_shares)


if __name__ == "__main__":
    main()



"""
You want to print data using print(), but you also want to change the separator
character or line ending.
"""


def main():
    print("ACME", 50, 91.5)
    print("ACME", 50, 91.5, sep=",")
    print("ACME", 50, 91.5, sep=",", end="!!\n")
    for i in range(5):
        print(i, end=" ")


if __name__ == "__main__":
    main()



"""
You're writing code that uses callback functions, but you're concerned
about the proliferation of small functions and mind-boggling control
flow. You would like some way to make the code look more like a sequence
of procedural steps.
"""
from queue import Queue
from functools import wraps


def apply_async(func, args, *, callback):
    # Compute the result
    result = func(*args)

    # Invoke the callback with the result
    callback(result)


def add(x, y):
    return x + y


def inlined_async(func):
    @wraps(func)
    def wrapper(*args):
        f = func(*args)
        result_queue = Queue()
        result_queue.put(None)

        while True:
            result = result_queue.get()
            try:
                a = f.send(result)
                apply_async(a.func, a.args, callback=result_queue.put)
            except StopIteration:
                break

    return wrapper


class Async:
    def __init__(self, func, args):
        self.func = func
        self.args = args


@inlined_async
def example_1():
    r = yield Async(add, (2, 3))
    print(r)
    r = yield Async(add, ("Hello ", "World!"))
    print(r)

    for n in range(10):
        r = yield Async(add, (n, n))
        print(r)
    print("Goodbye")


def main():
    example_1()


if __name__ == "__main__":
    main()



"""
You want to know in detail what your code is doing under the hood by
disassembling it into lower-level byte code used by the interpreter.
"""

import opcode


def generate_opcodes(codebytes):
    extended_arg = 0
    i = 0
    n = len(codebytes)
    while i < n:
        op = codebytes[i]
        i += 1
        if op >= opcode.HAVE_ARGUMENT:
            oparg = codebytes[i] + codebytes[i + 1] * 256 + extended_arg
            extended_arg = 0
            i += 2
            if op == opcode.EXTENDED_ARG:
                extended_arg = oparg * 65536
                continue
        else:
            oparg = None
        yield (op, oparg)


# Example
def countdown(n):
    while n > 0:
        print("T-minus", n)
        n -= 1
    print("Blastoff!")


def main():
    try:
        for op, oparg in generate_opcodes(countdown.__code__.co_code):
            print(op, opcode.opname[op], oparg)
    except Exception as exc:
        print(exc)


if __name__ == "__main__":
    main()



"""
You need to decode or encode binary data using Base64 encoding
"""
import base64


def main():
    s = b"hello"
    a = base64.b64encode(s)
    print("s:", s)
    print("a:", a)

    print("decoded:", base64.b64decode(a))


if __name__ == "__main__":
    main()



"""
You are using regular expressions to process text, but are
concerned about the handling of Unicode characters.
"""

import re


def main():
    num = re.compile(r"\d+")
    print("num:", num)
    print("match 123:", num.match("123"))
    print("match \u0661\u0662\u0663", num.match("123"))
    print()

    pat = re.compile("stra\u00dfe", re.IGNORECASE)
    s = "straße"

    print("pat:", pat)
    print("s:", s)
    print("s matches:", pat.match(s))
    print("s matches UPPER:", pat.match(s.upper()))
    print("s.upper():", s.upper())


if __name__ == "__main__":
    main()



"""
You want to write data to a file, but only if it doesn't already exist
on a filesystem.
"""

from pathlib import Path


def main():
    with (Path(__file__).parent / "somefile.txt").open("w+") as f:
        f.write("hello world")

    try:
        with (Path(__file__).parent / "somefile.txt").open("xt") as f:
            print(f.read())
    except FileExistsError:
        print("file already exists!")


if __name__ == "__main__":
    main()



"""
You want to implement a custom iteration pattern that's different
than the usual built-in functions (e.g. range(), reversed(), etc.)
"""


def frange(start, stop, increment):
    x = start
    while x < stop:
        yield x
        x += increment


def countdown(n):
    while n > 0:
        yield n
        n -= 1


def main():
    print("For loop - starting at 0, ending at 4 and incrementing by 0.5")
    for x in frange(0, 4, 0.5):
        print(x)

    print("\nList of valuee between 0 and 1 with a 0.125 increment")
    print(list(frange(0, 1, 0.125)))

    print("\nCounting down from 5")
    for x in countdown(5):
        print(x)
    print("Done!\n")

    print(f"list of values from the countdown: {list(countdown(5))}")


if __name__ == "__main__":
    main()



"""
You want to read or write data encoded as JSON (JavaScript Object Notation).
"""

import json
from pathlib import Path


def main():
    data = {"name": "AAPL", "shares": 10, "price": 100}
    print(data)

    json_str = json.dumps(data)
    data = json.loads(json_str)

    print(data)
    print(json_str)

    with open(Path(__file__).parent / "data.json", "r") as file:
        data = json.load(file)
    print(data)


if __name__ == "__main__":
    main()



"""
You want to take the data in a Python dictionary and turn it into XML.
"""
from xml.etree.ElementTree import Element, tostring


def dict_to_xml(tag, d):
    """Turn a simple dict of key/value pairs into XML"""
    elem = Element(tag)
    for key, val in d.items():
        child = Element(key)
        child.text = str(val)
        elem.append(child)
    return elem


def main():
    s = {"name": "GOOG", "shares": 100, "price": 490.1}
    print("s:", s)
    e = dict_to_xml("stock", s)
    print("e:", e)
    print("e to string:", tostring(e))

    # Set some attributes
    e.set("_id", "1234")
    print("e to string:", tostring(e))


if __name__ == "__main__":
    main()



"""
Your application receives temporal data in string format, but you
want to convert those strings into datetime objects to perform
non-string operations on them.
"""
from datetime import datetime


def main():
    text = "2012-09-20"
    y = datetime.strptime(text, "%Y-%m-%d")
    z = datetime.now()
    print("text:", text)
    print("y:", y)
    print("now:", z)
    print("now (text):", z.strftime("%A %B %d, %Y"))
    print("now - y:", z - y)


if __name__ == "__main__":
    main()



"""
You want to make a dictionary that is a subset of another dictinary.
"""

from pprint import pprint


def main():
    # example of extracting a subset from a dictionary
    prices = {"ACME": 45.23, "AAPL": 612.78, "IBM": 205.55, "HPQ": 37.20, "FB": 10.75}

    # Make a dictionary of all prices over 200
    p1 = {key: value for key, value in prices.items() if value > 200}

    print("All prices over 200")
    pprint(p1)

    # Make a dictionary of tech stocks
    tech_names = {"AAPL", "IBM", "HPQ", "MSFT"}
    p2 = {key: value for key, value in prices.items() if key in tech_names}

    print("All techs")
    pprint(p2)


if __name__ == "__main__":
    main()



"""
A decorator has been applied to a function, but you want to 'undo' it,
gaining access to the original unwrapped function.
"""
from functools import wraps
import time


def somedecorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        print(
            f'Invoking the function "{func.__name__}" with\n  args: {args}\n  kwargs: {kwargs}'
        )
        result = func(*args, **kwargs)
        print(f"Got result {result} in {time.time() - start: .5f} seconds")

    return wrapper


@somedecorator
def add(a: int, b: int):
    return a + b


def main():
    print("Running withthe decorator as normal:")
    result = add(2, 3)
    print(result)

    print("\nunwrapping to run original function:")
    result = add.__wrapped__(2, 3)
    print(result)


if __name__ == "__main__":
    main()



"""
You want to define a class that serves as an interface or abstract base
class from which you can perform type checking and ensure certain methods
are implemented in subclasses.
"""
from abc import ABCMeta, abstractmethod


class IStream(metaclass=ABCMeta):
    @abstractmethod
    def read(self, maxbytes=-1):
        pass

    @abstractmethod
    def write(self, data):
        pass


class Socket(IStream):
    def read(self, maxbytes=-1):
        ...

    def write(self, data):
        ...


def serialize(obj, stream):
    if not isinstance(stream, IStream):
        raise TypeError("Expected an IStream")
    print("obj is an IStream - serializing...")


def main():
    try:
        IStream()
    except TypeError as exc:
        print(exc)

    socket = Socket()
    print(socket)
    serialize("somefile.text", socket)


if __name__ == "__main__":
    main()



"""
You want to read or write data encoded as a binary array of uniform
structures into Python tuples.
"""
from struct import Struct
from pathlib import Path


def write_records(records, fmt, f):
    """Write a sequence of tuples to a binary file or structure"""
    record_struct = Struct(fmt)
    for r in records:
        f.write(record_struct.pack(*r))


def read_records(fmt, f):
    record_struct = Struct(fmt)
    chunks = iter(lambda: f.read(record_struct.size), b"")
    return (record_struct.unpack(chunk) for chunk in chunks)


def example_1():
    records = [(1, 2.3, 4.5), (6, 7.8, 9.0), (12, 13.4, 56.7)]
    print("records:", records)

    with open(Path(__file__).parent / "data.b", "wb") as f:
        write_records(records, "<idd", f)


def example_2():
    with open(Path(__file__).parent / "data.b", "rb") as f:
        for rec in read_records("<idd", f):
            print(rec)


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You need to create a temporary file or directory for use when your program
executes. Afterwards, you possible want the file or directory to be destroyed.
"""
from tempfile import TemporaryDirectory, TemporaryFile, NamedTemporaryFile
from pathlib import Path


def example_1():
    """Unnamed temp file"""
    with TemporaryFile("w+t") as f:
        # Read / Write to the file
        f.write("Hello World!\n")
        f.write("Testing\n")

        # Seek back to beginning and read
        f.seek(0)
        data = f.read()
        print(data)
    # Temp file is destoyed
    print()


def example_2():
    """Named temp file"""
    with NamedTemporaryFile("w+t") as f:
        temp_file_name = f.name
        print("tmp file:", temp_file_name)
        print("exists:", Path(temp_file_name).exists())
    print("Temp file destroyed")
    print("exists:", Path(temp_file_name).exists())
    print()


def example_3():
    """Temp directory"""
    with TemporaryDirectory() as dirname:
        print("tmp dirname:", dirname)
        print("exists:", Path(dirname).exists())
    print("tmp dir destroyed")
    print("exists:", Path(dirname).exists())
    print()


def main():
    example_1()
    example_2()
    example_3()


if __name__ == "__main__":
    main()



"""
You have a collection of sorted sequences and you want to iterate over a sorted
sequence of them all merged together .
"""

import heapq


def main():
    a = [1, 4, 7, 10]
    b = [2, 5, 6, 11]
    c = [17, 23, 43, 86]

    print(f"a: {a}")
    print(f"b: {b}")
    print(f"c: {c}")
    print("merging and iterating into single sorted output:")
    for x in heapq.merge(a, b, c):
        print(x)


if __name__ == "__main__":
    main()



"""
Your program creates a large number (e.g. millions) of instances
and uses a large amount of memory.
"""


class Date:
    __slots__ = ["year", "month", "day"]

    def __init__(self, year, month, day):
        self.year = year
        self.month = month
        self.day = day


def main():
    d = Date(2023, 1, 1)
    print(d)
    print(d.__slots__)
    print(d.year)
    print(d.month)
    print(d.day)


if __name__ == "__main__":
    main()



"""
You want to iterate over a sequence, but would like to keep track of which
element of the sequence is currently being processed.
"""


def main():
    my_list = ["a", "b", "c"]

    print(f"my_list: {my_list}")
    print("enumerating - default")
    for idx, val in enumerate(my_list):
        print(idx, val)

    print("enumerating - idx stars at 1")
    for idx, val in enumerate(my_list, 1):
        print(idx, val)


if __name__ == "__main__":
    main()



"""
You have a sequence of dictionaries or instances and you want to iterate over the
data in groups based on the value of a particular field, such as date.
"""

from operator import itemgetter
from itertools import groupby
from pprint import pprint


def main():
    rows = [
        {"address": "5412 N CLARK", "date": "07/01/2012"},
        {"address": "5148 N CLARK", "date": "07/04/2012"},
        {"address": "5800 E 58TH", "date": "07/02/2012"},
        {"address": "2122 N CLARK", "date": "07/03/2012"},
        {"address": "5645 N RAVENSWOOD", "date": "07/02/2012"},
        {"address": "1060 W ADDISON", "date": "07/02/2012"},
        {"address": "4801 N BROADWAY", "date": "07/01/2012"},
        {"address": "1039 W GRANVILLE", "date": "07/04/2012"},
    ]

    print("Original rows:")
    pprint(rows)

    # Sort first by the desired field
    rows.sort(key=itemgetter("date"))
    print("\nRows - sorted by date")
    pprint(rows)

    # Iterate over groups
    print("\nIterating over groups (grouped by date)")
    for date, items in groupby(rows, key=itemgetter("date")):
        print(f"Date: {date}")
        for item in items:
            print(f"  {item}")
        print()


if __name__ == "__main__":
    main()



"""
You have code that needs to perform simple time conversions, like
days to seconds, hours to minutes, and so on.
"""
from datetime import timedelta, datetime


def main():
    a = timedelta(days=2, hours=6)
    b = timedelta(hours=4.5)
    c = a + b
    print("a:", a)
    print("b:", b)
    print("c:", c)
    print("c days:", c.days)
    print("seconds:", c.seconds)
    print("total seconds:", c.total_seconds())
    print("hours:", c.total_seconds() / 3600)
    print()

    d = datetime(2023, 1, 1)
    now = datetime.now()
    print("d:", d)
    print("d + a:", d + a)
    print("now:", now)
    print("now + a:", now + a)


if __name__ == "__main__":
    main()



"""
You want to process data iteritively in the style of a data processing
pipeline (similar to Unix pipes). For instance, you have a huge amount
of data that needs to be processed but it can't entirely fit into memory.
"""

import os
import fnmatch
import gzip
import bz2
import re
from pathlib import Path


def gen_find(filepat, top):
    """
    Find all filenames in a directory tree that match a shell wildcard pattern.
    """
    for path, _, filelist in os.walk(top):
        for name in fnmatch.filter(filelist, filepat):
            yield os.path.join(path, name)


def gen_opener(filenames):
    """
    Open a sequence of filenames one at a time producing a file object.
    The file is closed immediately when proceeding to the next iteration.
    """
    for filename in filenames:
        if filename.endswith(".gz"):
            f = gzip.open(filename, "rt")
        elif filename.endswith(".bz2"):
            f = bz2.open(filename, "rt")
        else:
            f = open(filename, "rt")
        yield f
        f.close()


def gen_concatenate(iterators):
    """
    Chain a sequence of iterators together to a single sequence.
    """
    for iterator in iterators:
        yield from iterator


def gen_grep(pattern, lines):
    """
    Look for a regex pattern in a sequence of lines
    """
    pat = re.compile(pattern)
    for line in lines:
        if pat.search(line):
            yield line


def example_1():
    lognames = gen_find("access-log*", Path(__file__).parent / "data/")
    files = gen_opener(lognames)
    lines = gen_concatenate(files)
    pylines = gen_grep("(?i)python", lines)
    bytecolumns = (line.rsplit(None, 1)[1] for line in pylines)
    bytes_ = (int(x) for x in bytecolumns if x != "-")
    print(f"Bytes: {sum(bytes_):,}")


def example_2():
    total = 0
    lognames = gen_find("access-log*", Path(__file__).parent / "data/")
    files = gen_opener(lognames)
    lines = gen_concatenate(files)
    pylines = gen_grep("(?i)python", lines)
    total = (1 for _ in pylines)
    print(f"Lines: {sum(total):,}")


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You want to get a list of files contained in a directory on the filesystem.
"""
from pathlib import Path

somedir = Path(__file__).parent


def main():
    names = list(somedir.iterdir())
    print(names)

    # Get files
    print("\nfiles:")
    for name in names:
        if name.is_file():
            print(name)

    # Get dirs
    print("\ndirs:")
    for name in names:
        if name.is_dir():
            print(name)

    # Get specific files
    print("\n.py files")
    for name in names:
        if name.suffix == ".py":
            print(name)

    # Glob matching recursive
    print("\nglob matching pattern *.py")
    for name in somedir.glob("**/*.py"):
        print(name)


if __name__ == "__main__":
    main()



"""
You need to serialize a Python object into a byte stream so that
you can do things such as save it to a file, store it in a database,
or transmit it over a network connection.
"""
import pickle
from pathlib import Path
from fractions import Fraction
from decimal import Decimal


parent_dir = Path(__file__).parent


def main():
    # Some python object
    obj = [
        "A",
        "B",
        "C",
        (1, 2, 3),
        {1: [1, 2, 3.5], 2: ["a", "b", "c"]},
        Fraction(3, 16),
        Decimal("2.45"),
    ]
    print("obj:", obj)

    # "pickle" the object to a file
    with parent_dir.joinpath("pickled").open("wb") as f:
        pickle.dump(obj, f)

    # "pickle" to a string
    pickled_string = pickle.dumps(obj)
    print("pickle string:", pickled_string)

    # "unpickle" the object from a file
    with parent_dir.joinpath("pickled").open("rb") as f:
        unpickled = pickle.load(f)
        print("unpickeld file:", unpickled)

    # "unpickle" from string
    unpickled = pickle.loads(pickled_string)
    print("unpickled string:", unpickled)


if __name__ == "__main__":
    main()



"""
You'd like to define a read-only attribute as a property that only gets
computed on access. However, once accessed, you'd like the value to be
cached and not recomputed each access.
"""

import math


class lazyproperty:
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls):
        if instance is None:
            return self
        else:
            value = self.func(instance)
            setattr(instance, self.func.__name__, value)
            return value


class Circle:
    def __init__(self, radius):
        self.radius = radius

    @lazyproperty
    def area(self):
        print("Computing area...")
        return math.pi * self.radius**2

    @lazyproperty
    def perimeter(self):
        print("Calculating perimeter...")
        return 2 * math.pi * self.radius


def main():
    c = Circle(4.0)
    print(vars(c))

    print(c.radius)
    print(c.area)
    print(c.area)
    print(c.perimeter)
    print(c.perimeter)

    print(vars(c))


if __name__ == "__main__":
    main()



"""
You want a general solution for finding a date for the last
occurance of a day of the week. Last Friday, for example.
"""
from datetime import datetime, timedelta


weekdays = [
    "Monday",
    "Tuesday",
    "Wednesday",
    "Thursday",
    "Friday",
    "Saturday",
    "Sunday",
]


def get_previous_byday(dayname, start_date=None):
    if start_date is None:
        start_date = datetime.today()
    day_num = start_date.weekday()
    day_num_target = weekdays.index(dayname)
    days_ago = (7 + day_num - day_num_target) % 7
    if days_ago == 0:
        days_ago = 7
    return start_date - timedelta(days=days_ago)


def main():
    today = datetime.today()
    print("today:", today)
    print("last Friday:", get_previous_byday("Friday", today))
    print("last Monday:", get_previous_byday("Monday", today))
    print("last Tuesday:", get_previous_byday("Tuesday", today))

    date = datetime(2023, 6, 1, 12)
    print("date:", date)
    print("previous Friday:", get_previous_byday("Friday", date))
    print("previous Monday:", get_previous_byday("Monday", date))
    print("previous Tuesday:", get_previous_byday("Tuesday", date))


if __name__ == "__main__":
    main()



"""
You want to find the largest of smallest N items in a collection.
"""

import heapq
from pprint import pprint


def example_1():
    nums = [1, 8, 2, 23, 7, -4, 18, 23, 42, 37, 2]
    print(f"Nums: {nums}")
    print(f"Largest 3: {heapq.nlargest(3, nums)}")
    print(f"Smallest 3: {heapq.nsmallest(3, nums)}")


def example_2():
    portfolio = [
        {"name": "IBM", "shares": 100, "price": 91.1},
        {"name": "AAPL", "shares": 50, "price": 543.22},
        {"name": "FB", "shares": 200, "price": 21.09},
        {"name": "HPQ", "shares": 35, "price": 31.75},
        {"name": "YHOO", "shares": 45, "price": 16.35},
        {"name": "ACME", "shares": 75, "price": 115.65},
    ]
    print("portfolio")
    pprint(portfolio)
    print("\n3 cheapest stocks:")
    pprint(heapq.nsmallest(3, portfolio, key=lambda x: x["price"]))
    print("\n3 most expensive stocks:")
    pprint(heapq.nlargest(3, portfolio, key=lambda x: x["price"]))


def main():
    example_1()
    print("=" * 50)
    example_2()


if __name__ == "__main__":
    main()



from typing import Any


class ChapterNotFoundError(Exception):
    def __init__(self, value: Any):
        self.value = value
        super().__init__(f"Chapter {value} not found")


class RecipeNotFoundError(Exception):
    def __init__(self, value: Any):
        self.value = value
        super().__init__(f"Recipe {value} not found")



"""
You want to create and destroy threads for concurrent execution of code.
"""

import time
from threading import Thread


class CountdownTask:
    def __init__(self):
        self._running = True

    def terminate(self):
        self._running = False

    def run(self, n):
        while self._running and n > 0:
            print("T-minus", n)
            n -= 1
            time.sleep(1)


def countdown(n):
    while n > 0:
        print("T-minus", n)
        n -= 1
        time.sleep(1)


def main():
    t = Thread(target=countdown, args=(10,), daemon=True)
    print(t)
    t.start()
    if t.is_alive():
        print("t is still running")
    else:
        print("t is finished")

    c = CountdownTask()
    t2 = Thread(target=c.run, args=(10,), daemon=True)
    t2.start()


if __name__ == "__main__":
    main()



"""
You have a sequence of items and you'd like to determine the most frequently
occurring items in the sequence.
"""

from collections import Counter


def main():
    words = [
        "look",
        "into",
        "my",
        "eyes",
        "look",
        "into",
        "my",
        "eyes",
        "the",
        "eyes",
        "the",
        "eyes",
        "the",
        "eyes",
        "not",
        "around",
        "the",
        "eyes",
        "don't",
        "look",
        "around",
        "the",
        "eyes",
        "not",
        "around",
        "the",
        "eyes",
        "look",
        "into",
        "the",
        "eyes",
        "my",
        "eyes",
        "you'r",
        "under",
    ]

    word_counts = Counter(words)
    top_three = word_counts.most_common(3)

    print(f"\nwords: {words}", end="\n\n")
    print(f"word counter: {word_counts}", end="\n\n")
    print(f"top three: {top_three}")


if __name__ == "__main__":
    main()



"""
You would like to write a single decorator that can be used without arguments,
such as @decorator, or with optional arguments, such as @decorator(x, y, z).
However, there seems to be no straightforward way to do it due to differences
in calling conventions between simple decorators and decorators taking arguments.
"""

from functools import wraps, partial
import logging


def logged(func=None, *, level=logging.DEBUG, name=None, message=None):
    if func is None:
        return partial(logged, level=level, message=message)
    logname = name if name else func.__module__
    log = logging.getLogger(logname)
    logmsg = message if message else func.__name__

    @wraps(func)
    def wrapper(*args, **kwargs):
        log.log(level, logmsg)
        return func(*args, **kwargs)

    return wrapper


@logged
def add(x, y):
    return x + y


@logged(level=logging.CRITICAL, name="spam", message="UH OH! Theres SPAM!")
def spam():
    return "SPAM!"


def main():
    logging.basicConfig(level=logging.DEBUG)
    print(add(1, 2))
    print(spam())


if __name__ == "__main__":
    main()



"""
You want to perform common text operations (e.g. stripping, searching
and replacement) on byte strings.
"""

import re


def main():
    data = b"Hello World!"
    print("data:", data)
    print(data[:5])
    print(data.startswith(b"Hello"))
    print(data.split())
    print(data.replace(b"Hello", b"Hello Cruel"))
    print()

    data = bytearray(data)
    print("data:", data)
    print(data[:5])
    print(data.startswith(b"Hello"))
    print(data.split())
    print(data.replace(b"Hello", b"Hello Cruel"))
    print()

    data = b"FOO:BAR,SPAM"
    pattern = re.compile(b"[:,]")
    splits = re.split(pattern, data)
    print("data:", data)
    print("pattern:", pattern)
    print("splits:", splits)


if __name__ == "__main__":
    main()



"""
You want to iterate in reverse over a sequence.
"""


class Countdown:
    def __init__(self, start):
        self.start = start

    # Forward iterator
    def __iter__(self):
        n = self.start
        while n > 0:
            yield n
            n -= 1

    # Reverse iterator
    def __reversed__(self):
        n = 1
        while n <= self.start:
            yield n
            n += 1


def main():
    c = Countdown(5)
    print("Forward:")
    for x in c:
        print(x)

    print("Reverse:")
    for x in reversed(c):
        print(x)


if __name__ == "__main__":
    main()



"""
You are building custom objects on which you would like to support iteration,
but would like an easy way to implement the iterator protocol.
"""


class Node:
    def __init__(self, value):
        self.value = value
        self._children = []

    def __repr__(self):
        return f"Node({self.value})"

    def add_child(self, node):
        self._children.append(node)

    def __iter__(self):
        return iter(self._children)

    def depth_first(self):
        yield self
        for child in self:
            yield from child.depth_first()


def main():
    root = Node(0)
    child1 = Node(1)
    child2 = Node(2)
    child3 = Node(3)
    root.add_child(child1)
    root.add_child(child2)
    root.add_child(child3)
    child1.add_child(Node(3))
    child2.add_child(Node(4))
    child3.add_child(Node(5))
    child3.add_child(Node(6))

    for child in root.depth_first():
        print(child)

    print(list(iter(root)))


if __name__ == "__main__":
    main()



"""
You need to read or write binary data, such as that found in images,
sound files, and so on.
"""

from pathlib import Path


def main():
    with (Path(__file__).parent / "data.bin").open("wb") as f:
        f.write(b"hello world")

    with (Path(__file__).parent / "data.bin").open("rb") as f:
        print(f.read())


if __name__ == "__main__":
    main()
